/********************************* gps.c ***************************************
*                                                                              *
*  Author: David Collins                                                       *
*  Project: Global Pigeon - ECE 477 Team 10, Spring 2008                       *
*                                                                              *
*  Allows for communication with an ETek EB-85A GPS module.  However, the      *
*  usage of of standard NMEA sentences should allow this source code to work   *
*  with other GPS modules.                                                     *
*                                                                              *
*  Several buffers are maintained into which raw NMEA sentences are written.   *
*  Writing to a buffer begins when an '$' is received and stops when a '*' is  *
*  received.  Only chosen NMEA sentences are stored.  These consist of RMC,    *
*  GGA, and GSA.  A translation function is called to fill integer values in a *
*  structure based on the contents of each NMEA sentence.                      *
*                                                                              *
*******************************************************************************/


#include <hidef.h> 
#include <mc9s12a64.h>
#include "gps.h"
#include "sci.h"
#include "debug.h"
#include "nav.h"
// #include "constants.h"


const char NMEA_SET_MIN[] =   "$PMTK314,0,1,0,1,5,0,0,0,0,0,0,0,0,0,0,0,0*2D\r\n";
//const char NMEA_SET_MIN[] = "$PMTK314,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
//const char NMEA_SET_MIN[] = "$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
const char NMEA_SET_WAAS[] = "$PMTK313,1*2E\r\n$PMTK301,2*2E\r\n";


#define GPS_NUM_ACCEPTED 3

const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPRMC", "GPGGA", "GPGSA"};
//const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPVTG", "GPGGA"};
//const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPRMC", "GPGGA"};


#define GPS_NUM_BUFFERS 5
#define GPS_BUFFER_SIZE 255

// head and tail points for buffer selection
int gpsHead, gpsTail;

// GPS input buffer
char gpsBuffer[GPS_NUM_BUFFERS][GPS_BUFFER_SIZE];

// current writing position in each GPS buffer
int gpsBufferPos[GPS_NUM_BUFFERS];

// flag is true while currently receiving a NMEA sentence 
char gpsInSentence;
// flag true if NMEA sentence
char gpsInAccepted;



// bits for accepted sentences set high when received
int gpsFlag;


GPS_DATA gpsData, gpsDataStored;

// initial the GPS module and the UART port connecting to it
void GPS_Init(void) {
   int i;
   unsigned char *p;

   // initialize gps buffers to 0
   gpsHead = gpsTail = 0;
   for (i=0; i<GPS_NUM_BUFFERS; i++) {
      gpsBufferPos[i] = 0;
   }

   gpsInSentence = 0;
   gpsInAccepted = 0;
   gpsFlag = 0;

   // zero out the gps data structure
   p = (unsigned char *)&gpsData;
   for (i=0; i<sizeof(GPS_DATA); i++) {
      p[i] = 0;
   }

   // initialize the SCI0 registers
   SCI_Init(0, 38400);
   
   // set for receiving interrupt
   SCI0CR2 = 0x2C;
   
   
   // clear the flags
   //SCI0SR1 = 0xC0;			   
   
   // send NMEA command to set for minimal sentence output
   SCI_OutString(0, NMEA_SET_MIN);

   // send NMEA command to enable WAAS operation   
   SCI_OutString(0, NMEA_SET_WAAS);	
   SCI_OutString(1, "\r\nSent commands \r\n\r\n");			
   
}


// returns true when strings match
int IsMatch(const char *a, const char *b, int len) 
{
   int i;
   for (i=0; i<len; i++) 
   {
      if (a[i] != b[i]) return 0;
   }
   return 1;
}




// returns true if new NMEA sentence is ready for translation
unsigned char GPS_ReadyFlag(void) 
{
   return (gpsTail != gpsHead);
}



// handle incoming data using receiver interrupt
interrupt 20 void GPS_Receive_Handler(void) {
   int i;
   char c;
   
   // clear the flag
   i = SCI0SR1;
   // get the new character
   c = SCI0DRL;
   
   
   // check that it is a receive interrupt
   if (!(i & RDRF)) 
   {
      SCI_OutString(1, "SCI interrupted but not on receive.\r\n");
      // should only get hear through an interrupt,
      // so the flag ought always be set
      // return in its not
      return;
   }				  
   
   // pass through the character
#ifdef GPSPASSTHRU   
   SCI_OutChar(1, c);
#endif
   
  
   if (!gpsInSentence) {
      if (c == '$') {            
         gpsInSentence = 1;
      }
   }
   else if (!gpsInAccepted) {
      if (c == ',') {
         // TalkerID finished, check if it is accepted            
         for (i=0; i<GPS_NUM_ACCEPTED; i++) {               
            if (IsMatch(gpsBuffer[gpsTail], gpsNmeaAccepted[i], gpsBufferPos[gpsTail])) {
               // it is an accepted sentence
               break;
            }
         }
         if (i<GPS_NUM_ACCEPTED) {               
            // replace first character in buffer with accepted NMEA index
            gpsBuffer[gpsTail][0] = '0' + i;
            gpsBuffer[gpsTail][1] = ',';
            gpsBufferPos[gpsTail] = 2;
            gpsInAccepted = 1;
         }
         else {               
            // this is not an accepted sentence, so set for not in sentence
            gpsInSentence = 0;
            // move position back to start of buffer
            gpsBufferPos[gpsTail] = 0;
         }
      }
      else {
         // add character to buffer
         gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = c;
         gpsBufferPos[gpsTail]++;
      }
   }
   else {
      // is in an accepted NMEA sentence, copy into buffer if not end
               
      if (c == '*') {            
         // end of the sentence
         gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = '\0';
         gpsBufferPos[gpsTail]++;

         // set for not in sentence and not accepted
         gpsInSentence = 0;
         gpsInAccepted = 0;

         // move to next buffer with circular add
         gpsTail++;
         if (gpsTail == GPS_NUM_BUFFERS) gpsTail=0;            

         // move to the start of the buffer
         gpsBufferPos[gpsTail] = 0;
      }
      else {
         // just copy the char
         gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = c;
         gpsBufferPos[gpsTail]++;
      }
   }	     
}		  



// non-interrupt function for translating NMEA sentences into useful data
// returns true if a fully refreshed reading is available
int GPS_Translate(void) {
   int i, j, neg, max;
   char *buf;

   
   if (gpsTail != gpsHead) {
      // new data is ready in the buffer
      //printf("%s\n", gpsBuffer[gpsHead]);

      buf = gpsBuffer[gpsHead];
      max = gpsBufferPos[gpsHead];
      
      i = buf[0] - '0';
      if (i == 0) {
         // RMC

         // get time
         gpsData.hr = (buf[2] - '0')*10 + (buf[3] - '0');
         gpsData.min = (buf[4] - '0')*10 + (buf[5] - '0');
         gpsData.sec = ((((buf[6]-'0')*10+(buf[7]-'0'))*10+(buf[9]-'0'))*10+(buf[10]-'0'))*10+(buf[11]-'0');
         
         gpsData.latitude = ((buf[15]-'0')*10+(buf[16]-'0'))*6;         
         // read in latitude
         for (i=17; i<24; i++) {
            if (buf[i] >= '0' && buf[i] <= '9') {
               if (i>17) gpsData.latitude *= 10;
               gpsData.latitude += buf[i] - '0';               
            }
         }
         if (buf[25] == 'S') {
            // make negative for southern hemisphere
            gpsData.latitude = 0 - gpsData.latitude;            
         }

         gpsData.longitude = (((buf[27]-'0')*10+(buf[28]-'0'))*10+(buf[29]-'0'))*6;
         // read in latitude
         for (i=30; i<37; i++) {
            if (buf[i] >= '0' && buf[i] <= '9') {
               if (i>30) gpsData.longitude *= 10;
               gpsData.longitude += buf[i] - '0';
            }
         }
         if (buf[38] == 'W') {
            // make negative for west
            gpsData.longitude = 0 - gpsData.longitude;
         }

         // get speed
         gpsData.speed = 0;
         i=40;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.speed *= 10;
               gpsData.speed += buf[i] - '0';
            }
            i++;
         }

         // get bearing
         gpsData.bearing = 0;
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.bearing *= 10;
               gpsData.bearing += buf[i] - '0';
            }
            i++;
         }

         // get date
         i++;
         gpsData.day = (buf[i]-'0')*10+(buf[i+1]-'0');
         i+=2;
         gpsData.month = (buf[i]-'0')*10+(buf[i+1]-'0');
         i+=2;
         gpsData.year = (buf[i]-'0')*10+(buf[i+1]-'0');
         
         //
         i+=5;
         switch (buf[i]) {
            case 'N':
               gpsData.navMode = 0;
               break;
            case 'A':
               gpsData.navMode = 1;
               break;
            case 'D':
               gpsData.navMode = 2;
               break;
            case 'E':
               gpsData.navMode = 3;
               break;
         }
         
         // ensure that the visible GPS value is always valid 
         gpsDataStored = gpsData;
         
         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPRMC;
         
      }
      else if (i == 1) {
         // GGA

         // get number of satellites
         gpsData.numSV = 0;
         i=40;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.numSV *= 10;
               gpsData.numSV += buf[i] - '0';
            }
            i++;
         }


         // altitude
         gpsData.altitude = 0;
         i++;
         j=0;
         while (j<1) {
            if (buf[i] == ',') j++;
            i++;
         }

         neg = 0;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.altitude *= 10;
               gpsData.altitude += buf[i] - '0';
            }
            else if (buf[i] == '-') {
               neg = 1;
            }
            i++;
         }
         // negate if needed
         if (neg) gpsData.altitude = - gpsData.altitude;



         // get geoidal separation
         gpsData.geoidSep = 0;
         i+=3;         
         neg = 0;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.geoidSep *= 10;
               gpsData.geoidSep += buf[i] - '0';
            }
            else if (buf[i] == '-') {
               neg = 1;
            }
            i++;
         }
         // negate if needed
         if (neg) gpsData.geoidSep = 0 - gpsData.geoidSep;

         // ensure that the visible GPS value is always valid 
         gpsDataStored = gpsData;

         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPGGA;
         
      }
      else if (i == 2) {
         // GSA

         // get fix mode
         gpsData.fix = buf[4] - '0';
         if (gpsData.fix == 1) gpsData.fix=0;

         // move to the DOPs
         i=6;
         j=0;
         while (j<12) {
            if (buf[i] == ',' && i < max) j++;
            i++;
         }

         // get PDOP
         gpsData.PDOP = 0;         
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.PDOP *= 10;
               gpsData.PDOP += buf[i] - '0';
            }
            i++;
         }

         // get PDOP
         gpsData.HDOP = 0; 
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.HDOP *= 10;
               gpsData.HDOP += buf[i] - '0';
            }
            i++;
         }

         // get PDOP
         gpsData.VDOP = 0; 
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.VDOP *= 10;
               gpsData.VDOP += buf[i] - '0';
            }
            i++;
         }
         
         // ensure that the visible GPS value is always valid 
         gpsDataStored = gpsData;

         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPGSA;         

      }



      // move to next buffer with circular add
      gpsHead++;
      if (gpsHead == GPS_NUM_BUFFERS) gpsHead=0;
   }


   // print everything when both are new
   if ((gpsFlag & (GPS_NMEA_GPRMC | GPS_NMEA_GPGGA)) == (GPS_NMEA_GPRMC | GPS_NMEA_GPGGA)) {
      gpsFlag &= ~(GPS_NMEA_GPRMC | GPS_NMEA_GPGGA);
   
      // print all data
		GPS_Print(&gpsData);      
     
      return 1;
   }
   
   return 0;
}

// copy the value of the internal gps structure into provided pointer
void GPS_Value(GPS_DATA *gps) 
{
   *gps = gpsDataStored;
}

// print out GPS data
void GPS_Print(const GPS_DATA *gps)
{
#ifndef GPSPASSTHRU
   SCI_OutChar(1, '\f');       


   SCI_OutLInt(1, gps->latitude / 6L); SCI_OutString(1, ", "); SCI_OutLInt(1, gps->longitude / 6L); SCI_OutString(1, "\r\n"); 
   SCI_OutULInt(1, gps->altitude); SCI_OutString(1, "m ");
   SCI_OutString(1, "SV:"); SCI_OutUInt(1, gps->numSV); SCI_OutString(1, " ");
   SCI_OutUInt(1, gps->fix); SCI_OutString(1, "D\r\n");
   SCI_OutUInt(1, gps->speed); SCI_OutString(1, "k "); 
   SCI_OutUInt(1, gps->bearing); SCI_OutString(1, "d\r\n");
   SCI_OutUInt(1, gps->hr); SCI_OutChar(1, ':'); SCI_OutUInt(1, gps->min); SCI_OutChar(1, ':'); SCI_OutULInt(1, gps->sec); SCI_OutString(1, "  "); 
   SCI_OutUInt(1, gps->month); SCI_OutChar(1, '/'); SCI_OutUInt(1, gps->day); SCI_OutChar(1, '/'); SCI_OutUInt(1, gps->year); SCI_OutString(1, "\r\n"); 
#endif      
      
      
   //SCI_OutString(1, "Nav mode - "); SCI_OutUInt(1, gps->navMode); SCI_OutString(1, "\r\n");       
   //SCI_OutString(1, "Geo Sep - "); SCI_OutUInt(1, gps->geoidSep); SCI_OutString(1, "\r\n");       


   //SCI_OutUInt(1, gps->PDOP); SCI_OutString(1, ", ");
   //SCI_OutUInt(1, gps->HDOP); SCI_OutString(1, ", ");
   //SCI_OutUInt(1, gps->VDOP); SCI_OutString(1, "\r\n");
}


// print out GPS data relative to a target
void GPS_PrintTarget(const GPS_DATA *gpsCurrent, const GPS_DATA *gpsTarget)
{
   unsigned long dist, bearing;
   long error;
   
   Nav_CalcDifference(&dist, &bearing, gpsCurrent, gpsTarget);

   error = (long)gpsCurrent->bearing / 100 - (long)bearing;
   if (error < -180) error += 360;
   else if (error > 180) error -=360;

   SCI_OutChar(1, '\f');
   SCI_OutLInt(1, gpsCurrent->latitude / 6L); SCI_OutString(1, ", "); SCI_OutLInt(1, gpsCurrent->longitude / 6L); SCI_OutString(1, "\r\n"); 
   SCI_OutULInt(1, gpsCurrent->altitude); SCI_OutString(1, "m ");
   SCI_OutString(1, "SV:"); SCI_OutUInt(1, gpsCurrent->numSV); SCI_OutString(1, " ");
   SCI_OutUInt(1, gpsCurrent->fix); SCI_OutString(1, "D\r\n");
   SCI_OutUInt(1, gpsCurrent->speed); SCI_OutString(1, "k "); 
   SCI_OutUInt(1, gpsCurrent->bearing); SCI_OutString(1, "d\r\n");
   SCI_OutULInt(1, dist); SCI_OutString(1, "ft ");
   SCI_OutULInt(1, bearing); SCI_OutString(1, "d ");
   SCI_OutLInt(1, error); SCI_OutString(1, "d"); 
}

/******************************** End of gps.c ********************************/
