/******************************* main.c ****************************************
*																										 *
*  Author: David Collins																		 *
*  Project: Global Pigeon - ECE 477 Team 10, Spring 2008								 *
*																										 *
*  This file contains the main function for Project Global Pigeon as well as	 *
*  many test functions which test various peripherials.  The code is designed	 *
*  to work with a Freescale 9S12A64 microcontroller as it to be compiled using *
*  CodeWarrior.																					 *
*																										 *
*  The current state of the main function consists of an infinite loop which	 *
*  waits for data and stores it periodically to the MicroSD card.  An RTI		 *
*  interrupt in nav.c maintains a consistent period and performs the actual    *
*  storage.  GPS coordinates, thermopile readings, and radio inputs are all	 *
*  logged.  This data was then going to be used to determine an appropriate	 *
*  control algorithm for the autopilot mode.  Unfortunately, poor weather,		 *
*  radio interference, and time constraints kept the autopilot mode from being *
*  realized.																						 *
*																										 *
*******************************************************************************/
 

#include <hidef.h>      // common defines and macros
#include <mc9s12a64.h>  // derivative information

#include "pll.h"
#include "timer.h"
#include "sci.h"
#include "gps.h"
#include "sd.h"
#include "pigeon.h"
#include "radio.h"
#include "servo.h"
#include "camera.h"	 
#include "adc.h"
#include "debug.h"
#include "nav.h"
#include "lcd.h"



#pragma LINK_INFO DERIVATIVE "mc9s12a64"

// debug testing for the GPS module
void TestGPS(void);

// output the binary offsets of the GPS data type
void PrintGPSOffset(void);

// simple debug testing of the MicroSD card
void TestSD(void);

// Store GPS track to the SD card
void TestTracking(void);

// display the radio values out of each channel
void TestRadio(void);

// interactive camera test function
void TestCamera(void);

// blink the LEDs
void TestHeartbeat(void);

// ensure that the thermopiles are outputting something reasonable
void TestThermopiles(void);

// test that distance and bearing calculation work properly
void Test_Bearings(void);

// display a menu for PSSC checkoff
void Test_PSSC(void);

// play with the servos
void PSSC3(void);

// play with MicroSD
void PSSC5(void);

// Store GPS track to the SD card while displaying distance and direction to engineering fountain
void TestTarget(void);

// Keep updating the GPS values
void TestGPSmonitor(void);

// display a menu to test the functionality of various peripherals
void TestMenu(void);



unsigned char SD_Buffer[SD_BLOCK_SIZE];

void PrintTime(unsigned long t_s, unsigned long t_e) 
{
   t_e -= t_s;
   SCI_OutString(1, "Time: ");
   SCI_OutULInt(1, t_e);
   SCI_OutString(1, "\r\n");
}







void main(void)
{       
   Pigeon_Init();
   PLL_Init();
   Timer_Init(0,1);
   
   //SCI_Init(1, 115200); // PC connection 
   LCD_Init(0); // LCD  
   
   GPS_Init();
   SD_Init();
   Radio_Init();
   Servo_Init();
   Camera_Init();
   ADC_Init();
      
   EnableInterrupts;       
   
   
   SCI_OutString(1, "Hello\r\n");     
   
   
   
   LEDS = LED_RED | LED_GREEN | LED_YELLOW;    
   
   //Test_Bearings();	    
   
   // spin forever, RTI in control
   //for(;;);     
   
   TestHeartbeat();        
      
   // turn off all LEDs   
   PORTA = 0;
   			     
   // start stabilization RTI,  needs to be after a heartbeat pause
   Nav_Init();
   
   // display GPS data along with distance and direction to a target
   //TestTarget();
   
   // PSSC checkoff menu 
   //Test_PSSC();
   
   // begins logging GPS waypoints immediately to the MicroSD card   
   //TestTracking();
   
   // Keep updating the GPS values while RTI is storing to the SD card
   TestGPSmonitor(); 
   
   
   // display a menu to test the functionality of various peripherals
   //TestMenu();
   
   // spin forever
   for (;;);
}


// ensure that the thermopiles are outputting something reasonable
void TestThermopiles(void)
{
   for(;;)
   {
      SCI_OutChar(1, '\f');
      SCI_OutString(1, "Vsense: "); SCI_OutUInt(1, ADC_In(ADC_CH_VSENSE)); SCI_OutString(1, "\r\n");
      SCI_OutString(1, "Therm0: "); SCI_OutUInt(1, ADC_In(ADC_CH_THERM0)); SCI_OutString(1, "\r\n");
      SCI_OutString(1, "Therm1: "); SCI_OutUInt(1, ADC_In(ADC_CH_THERM1)); SCI_OutString(1, "\r\n");
      SCI_OutString(1, "Therm2: "); SCI_OutUInt(1, ADC_In(ADC_CH_THERM2)); SCI_OutString(1, "\r\n");
      SCI_OutString(1, "Therm3: "); SCI_OutUInt(1, ADC_In(ADC_CH_THERM3)); SCI_OutString(1, "\r\n");
           
      Timer_Wait1ms(50);
   }   
}


// blink the LEDs
void TestHeartbeat(void)
{
   unsigned char i=0;
   int j;
   
   for (j=0; j<20; j++)
   //while (!SCI1_DATA_READY)
   {
      PORTA = i;
      PORTA |= 0b00000100;
      i++;
      
      Timer_Wait1ms(50);
   }
}


// interactive camera test function
void TestCamera(void)
{
   unsigned char keyPress=0;
   
   SCI_OutString(1, "On ->          O\r\n");
   SCI_OutString(1, "Off ->         P\r\n");
   SCI_OutString(1, "Auto focus ->  A\r\n");
   SCI_OutString(1, "Shutter ->     S\r\n");
   SCI_OutString(1, "Video start -> V\r\n");
   SCI_OutString(1, "Video stop ->  B\r\n");
   
   
   for(;;)
   {    
   
      // Test R707 Camera interface
      if (SCI1_DATA_READY)
      {	          
         keyPress = SCI_InChar(1);
         keyPress &= ~0x20; // convert to uppercase         
         
         if (keyPress == 'O') Camera_Send(CAMERA_COMMAND_ON);
         else if (keyPress == 'P') Camera_Send(CAMERA_COMMAND_OFF);
         else if (keyPress == 'A') Camera_Send(CAMERA_COMMAND_AF);
         else if (keyPress == 'S') Camera_Send(CAMERA_COMMAND_SHUTTER);
         else if (keyPress == 'V') Camera_Send(CAMERA_COMMAND_VIDEO_START);
         else if (keyPress == 'B') Camera_Send(CAMERA_COMMAND_VIDEO_STOP);
         else if (keyPress == 'Z') Camera_Send(CAMERA_COMMAND_AF_STROBE);
         else if (keyPress == 'X') Camera_Send(CAMERA_COMMAND_NULL_STROBE);
      }
    
      Camera_HandleEvents();
   }
   
}



// display the radio values out of each channel
void TestRadio(void)
{
   int i=0;
   
   for (;;) 
   {
      if ((i & 7) == 0)
      {
         // output reading
         SCI_OutChar(1, '\f');
         SCI_OutString(1, "Aux:      "); SCI_OutUInt(1, Radio_PulseWidth(0)); SCI_OutString(1, "\r\n");
         SCI_OutString(1, "Auto:     "); SCI_OutUInt(1, Radio_PulseWidth(1)); SCI_OutString(1, "\r\n");
         SCI_OutString(1, "Elevator: "); SCI_OutUInt(1, Radio_PulseWidth(2)); SCI_OutString(1, "\r\n");
         SCI_OutString(1, "Rudder:   "); SCI_OutUInt(1, Radio_PulseWidth(3)); SCI_OutString(1, "\r\n");
         SCI_OutString(1, "Ailerons: "); SCI_OutUInt(1, Radio_PulseWidth(4)); SCI_OutString(1, "\r\n");
         SCI_OutString(1, "Throttle: "); SCI_OutUInt(1, Radio_PulseWidth(5)); SCI_OutString(1, "\r\n");
      }
      i++;
      
      if (Radio_PulseWidth(RADIO_CH_AUTO) < 36000)
      {
         // auto mode
         
         // turn off pass through
         Radio_PassthruUnset(RADIO_MASK_ALL);
         
         Servo_Duty(SERVO_CH_THROTTLE, SERVO_IN_MIN);
         Servo_Duty(SERVO_CH_AILERON, 0);
         Servo_Duty(SERVO_CH_ELEVATOR, 0);
         Servo_Duty(SERVO_CH_RUDDER, 0);
      }
      else
      {
         // manual mode         
         Radio_PassthruSet(RADIO_MASK_ALL);
      }
      
      Timer_Wait1ms(10);
   }
}



// Store GPS track to the SD card
void TestTracking(void) 
{	    
   unsigned short i;
   unsigned long count = 0;   
   int max, bufcount=0;	 
    
   LEDS |= LED_RED;	  
   i = SD_Card_Init();
   if (i==0) LEDS &= ~LED_RED;
   
   // determine how many points will fit in a block
   // max = the max index
   max = (SD_BLOCK_SIZE / sizeof(GPS_DATA)) - 1;
   
   
     
   for (i=0; i<SD_BLOCK_SIZE; i++) SD_Buffer[i] = 0;
   // clear the header
   // i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   
   // read the header
   LEDS |= LED_RED;
   i = SD_Read_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   if (i==0) LEDS &= ~LED_RED;
   
   // copy out the first 4 bytes as an integer
   count = *(unsigned long *)SD_Buffer;
   
   for (;;)
   {	
      // new gps data ready
      if (GPS_ReadyFlag()) 
      {
      
         // translate the NMEA sentences to binary
         LEDS |= LED_YELLOW;
         if (GPS_Translate()) 
         {
            // only record it if all of the position and bearing fields refreshed, and 3D   
                        
            // copy the waypoint data into the SD buffer
            GPS_Value(&((GPS_DATA *)SD_Buffer)[bufcount]);
            
            //if (((GPS_DATA *)SD_Buffer)[bufcount].fix == 3) 
            // always plot
            if (1)
            {
               // 3D fix
               
               LEDS |= LED_GREEN;
               
               bufcount++;
               
               if (bufcount >= max)
               {
               #ifndef GPSPASSTHRU
                  //SCI_OutString(1, "Write "); 
                  //SCI_OutULInt(1, count);
                  //SCI_OutString(1, "\r\n");
               #endif
               
                  bufcount = 0;
                  
                  // write the data to the SD card
                  count++;
                  LEDS |= LED_RED;
                  i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT + count);
                  if (i==0) LEDS &= ~LED_RED;
                  
                  // write the count back to the header
                  *(unsigned long *)SD_Buffer = count;
                  LEDS |= LED_RED;
                  i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
                  if (i==0) LEDS &= ~LED_RED;
               }
            }
            else 
            {
               // Not a 3D fix
               
               LEDS &= ~LED_GREEN;
            }
         }
         LEDS &= ~LED_YELLOW;
         
      }     
   }
   
}


// Store GPS track to the SD card while displaying distance and direction to engineering fountain
void TestTarget(void) 
{	    
   unsigned short i;
   unsigned long count = 0;   
   int max, bufcount=0;	
   GPS_DATA gpsCurrent, gpsTarget;
   
   
   // set the engineering fountain as the target
   gpsTarget.latitude = 24257186L;
   gpsTarget.longitude = -52148269L;
   gpsTarget.altitude = 188;
    
   LEDS |= LED_RED;	  
   i = SD_Card_Init();
   if (i==0) LEDS &= ~LED_RED;
   
   // determine how many points will fit in a block
   // max = the max index
   max = (SD_BLOCK_SIZE / sizeof(GPS_DATA)) - 1;
   
   
     
   for (i=0; i<SD_BLOCK_SIZE; i++) SD_Buffer[i] = 0;
   // clear the header
   // i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   
   // read the header
   LEDS |= LED_RED;
   i = SD_Read_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   if (i==0) LEDS &= ~LED_RED;
   
   // copy out the first 4 bytes as an integer
   count = *(unsigned long *)SD_Buffer;
   
   for (;;)
   {	
      // new gps data ready
      if (GPS_ReadyFlag()) 
      {
      
         // translate the NMEA sentences to binary
         LEDS |= LED_YELLOW;
         if (GPS_Translate()) 
         {
            // only record it if all of the position and bearing fields refreshed, and 3D   
                        
            // copy the waypoint data into the SD buffer
            GPS_Value(&((GPS_DATA *)SD_Buffer)[bufcount]);
            GPS_Value(&gpsCurrent);
            
            //if (((GPS_DATA *)SD_Buffer)[bufcount].fix == 3) 
            // always plot
            if (1)
            {
               // 3D fix                 
                              
               LEDS |= LED_GREEN;
               
               GPS_PrintTarget(&gpsCurrent, &gpsTarget);
               
               bufcount++;
               
               if (bufcount >= max)
               {
               #ifndef GPSPASSTHRU
                  SCI_OutString(1, "Write "); 
                  SCI_OutULInt(1, count);
                  SCI_OutString(1, "\r\n");
               #endif
               
                  bufcount = 0;
                  
                  // write the data to the SD card
                  count++;
                  LEDS |= LED_RED;
                  i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT + count);
                  if (i==0) LEDS &= ~LED_RED;
                  
                  // write the count back to the header
                  *(unsigned long *)SD_Buffer = count;
                  LEDS |= LED_RED;
                  i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
                  if (i==0) LEDS &= ~LED_RED;
               }
            }
            else 
            {
               // Not a 3D fix
               
               LEDS &= ~LED_GREEN;
            }
         }
         LEDS &= ~LED_YELLOW;
         
      }     
   }
   
}




// debug testing for the GPS module
void TestGPS(void) 
{
   int i;
   unsigned char c;
   
   for(;;) 
   {     
      
      // wait for a byte from the GPS module
      /*
      while (!SCI0_DATA_READY);
      c = SCI_InChar(0);
      SCI_OutChar(1, c);
      */
      
      // call handler as if triggered by an interrupt
      // GPS_Receive_Handler();
      
      
      if (GPS_ReadyFlag()) 
      {
         i = GPS_Translate();
      } 
                            
   }   
}

// output the binary offsets of the GPS data type
// used to help translate from the codewarrior compiler to gcc storage mechanism
void PrintGPSOffset(void)
{
   GPS_DATA g;

   SCI_OutString(1, "sizeof(GPS_DATA) = "); SCI_OutUInt(1, sizeof(GPS_DATA)); SCI_OutString(1, "\r\nOffsets:\r\n");
   SCI_OutString(1, "latitude: ");  SCI_OutULInt(1, (unsigned long)&(g.latitude) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "longitude: ");  SCI_OutULInt(1, (unsigned long)&(g.longitude) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "altitude: ");  SCI_OutULInt(1, (unsigned long)&(g.altitude) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "speed: ");  SCI_OutULInt(1, (unsigned long)&(g.speed) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "bearing: ");  SCI_OutULInt(1, (unsigned long)&(g.bearing) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "numSV: ");  SCI_OutULInt(1, (unsigned long)&(g.numSV) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "year: ");  SCI_OutULInt(1, (unsigned long)&(g.year) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "month: ");  SCI_OutULInt(1, (unsigned long)&(g.day) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "day: ");  SCI_OutULInt(1, (unsigned long)&(g.day) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "hr: ");  SCI_OutULInt(1, (unsigned long)&(g.hr) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "min: ");  SCI_OutULInt(1, (unsigned long)&(g.min) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "sec: ");  SCI_OutULInt(1, (unsigned long)&(g.sec) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "geoidSep: ");  SCI_OutULInt(1, (unsigned long)&(g.geoidSep) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "PDOP: ");  SCI_OutULInt(1, (unsigned long)&(g.PDOP) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "HDOP: ");  SCI_OutULInt(1, (unsigned long)&(g.HDOP) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "VDOP: ");  SCI_OutULInt(1, (unsigned long)&(g.VDOP) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "fix: ");  SCI_OutULInt(1, (unsigned long)&(g.fix) - (unsigned long)&g); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "navMode: ");  SCI_OutULInt(1, (unsigned long)&(g.navMode) - (unsigned long)&g); SCI_OutString(1, "\r\n");
}


// simple debug testing of the MicroSD card
void TestSD(void) 
{	    
   unsigned short i,j;	
  
   char *replace = "OMG, h4x0red!!!";
   char *pos;
   
   // for timing purposes
   unsigned long t[4];
   
     
   SCI_OutString(1, "Initializing SD card...\r\n");
   t[0] = TimeStamp();
   i = SD_Card_Init();
   t[1] = TimeStamp();
   PrintTime(t[0], t[1]);
   
  
   
   SCI_OutString(1, "Reading block 645\r\n");
   t[0] = TimeStamp();
   i = SD_Read_Block(SD_Buffer, 645);   
   t[1] = TimeStamp();
   PrintTime(t[0], t[1]);
   
   SCI_OutString(1, "Before: ");
   for (i=0; i<30; i++)
   {
      SCI_OutChar(1, SD_Buffer[i]);
   }
   SCI_OutString(1, "\r\n");
   
   // modify the buffer
   pos = replace;
   i = 6;
   while (*replace) 
   {
      SD_Buffer[i] = *replace;
      replace++;
      i++;
   }  
   
   SCI_OutString(1, "After:  ");
   for (i=0; i<30; i++)
   {
      SCI_OutChar(1, SD_Buffer[i]);
   }
   SCI_OutString(1, "\r\n");
   
   //for (;;);
      
   
   SCI_OutString(1, "Writing edited block 645\r\n");  
   t[0] = TimeStamp();  
   i = SD_Write_Block(SD_Buffer, 645);
   t[1] = TimeStamp();
   PrintTime(t[0], t[1]);
   
   
   for (j=645; j<1000; j++)
   {	
      t[0] = TimeStamp();
      i = SD_Read_Block(SD_Buffer, j);
      t[1] = TimeStamp();
      PrintTime(t[0], t[1]);
           		      
      SCI_OutString(1, "Contents of memory block ");
      SCI_OutUInt(1, j);
      SCI_OutString(1, "\r\n");
      
      // print out a hex dump
      for (i=1; i<=SD_BLOCK_SIZE; i++)
      {
         SCI_OutHex(1, SD_Buffer[i-1]);
         if ( (i%16) == 0 )
            SCI_OutString(1, "\r\n");
         else if ( (i%8) == 0 )
            SCI_OutString(1, "  ");
         else
            SCI_OutChar(1, ' ');
      }
      SCI_OutString(1, "\r\n");
      SCI_OutString(1, "\r\n");
      
      // wait for key press
      i = SCI_InChar(1);
            
   }
}

// test that distance and bearing calculation work properly
void Test_Bearings(void) {
   unsigned long dist, bearing, truedist, truebearing;
   GPS_DATA p1, p2;

   SCI_OutString(1, "Distance and Bearing Test Cases:\r\n\r\n");


   p1.latitude = 24279031L;
   p1.longitude = -52154981L;
   p2.latitude = 24280528L;
   p2.longitude = -52155932L;
   truedist = 1010;
   truebearing = 334;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   SCI_OutString(1, "True dist: "); SCI_OutULInt(1, truedist); SCI_OutString(1, ", Calc dist: "); SCI_OutULInt(1, dist); SCI_OutString(1, "\r\n"); 
   SCI_OutString(1, "True bearing: "); SCI_OutULInt(1, truebearing); SCI_OutString(1, ", Calc bearing: "); SCI_OutULInt(1, bearing); SCI_OutString(1, "\r\n\r\n"); 
    
   truebearing = 154;
   
   Nav_CalcDifference(&dist, &bearing, &p2, &p1);
   
   SCI_OutString(1, "True dist: "); SCI_OutULInt(1, truedist); SCI_OutString(1, ", Calc dist: "); SCI_OutULInt(1, dist); SCI_OutString(1, "\r\n"); 
   SCI_OutString(1, "True bearing: "); SCI_OutULInt(1, truebearing); SCI_OutString(1, ", Calc bearing: "); SCI_OutULInt(1, bearing); SCI_OutString(1, "\r\n\r\n"); 
   

   p1.latitude = 24280008L;
   p1.longitude = -52155748L;
   p2.latitude = 24280029L;
   p2.longitude = -52155247L;
   truedist = 233;
   truebearing = 86;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   SCI_OutString(1, "True dist: "); SCI_OutULInt(1, truedist); SCI_OutString(1, ", Calc dist: "); SCI_OutULInt(1, dist); SCI_OutString(1, "\r\n"); 
   SCI_OutString(1, "True bearing: "); SCI_OutULInt(1, truebearing); SCI_OutString(1, ", Calc bearing: "); SCI_OutULInt(1, bearing); SCI_OutString(1, "\r\n\r\n"); 
   
   truebearing = 266;
   
   Nav_CalcDifference(&dist, &bearing, &p2, &p1);

   SCI_OutString(1, "True dist: "); SCI_OutULInt(1, truedist); SCI_OutString(1, ", Calc dist: "); SCI_OutULInt(1, dist); SCI_OutString(1, "\r\n"); 
   SCI_OutString(1, "True bearing: "); SCI_OutULInt(1, truebearing); SCI_OutString(1, ", Calc bearing: "); SCI_OutULInt(1, bearing); SCI_OutString(1, "\r\n\r\n"); 
   

   p1.latitude = 24256028L;
   p1.longitude = -52141495L;
   p2.latitude = 24255385L;
   p2.longitude = -52154478L;
   truedist = 6036;
   truebearing = 267;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   SCI_OutString(1, "True dist: "); SCI_OutULInt(1, truedist); SCI_OutString(1, ", Calc dist: "); SCI_OutULInt(1, dist); SCI_OutString(1, "\r\n"); 
   SCI_OutString(1, "True bearing: "); SCI_OutULInt(1, truebearing); SCI_OutString(1, ", Calc bearing: "); SCI_OutULInt(1, bearing); SCI_OutString(1, "\r\n\r\n"); 
   

   for(;;);
   
}

// display a menu for PSSC checkoff
void Test_PSSC(void) 
{
   unsigned char c;
   
   // choose a test program to run
   for (;;)
   {
      SCI_OutString(1, "\r\nChoose PSSC to test:\r\n");
      SCI_OutString(1, "2. An ability to decode and display GPS coordinates.\r\n");
      SCI_OutString(1, "3. An ability to control motor speed and control surface position.\r\n");
      SCI_OutString(1, "4. An ability to capture still images.\r\n");
      SCI_OutString(1, "5. An ability to read waypoints from a MicroSD removable storage device.\r\n");
      c = SCI_InChar(1);
      
      switch (c) 
      {
         case '2':    
            SCI_OutString(1, "\r\nPSSC 2\r\n");                         
            
            Timer_Wait1ms(200);
            
            SCI_Init(1, 19200);
            TestTracking(); 
            
            SCI_Init(1, 115200);           
            break;
         case '3':
            SCI_OutString(1, "\r\nPSSC 3\r\n");
            PSSC3(); 
            break;
         case '4':
            SCI_OutString(1, "\r\nPSSC 4\r\n");
            TestCamera();
            break;
         case '5':
            SCI_OutString(1, "\r\nPSSC 5\r\n");
            PSSC5();
            break;         
         default:
            break;
      }
   }
   
}


// play with the servos
void PSSC3(void)
{
   unsigned char c;
   
   
   // turn off radio pass thru
   Radio_PassthruUnset(RADIO_MASK_ALL);
   SET_AUTOMATIC_FLIGHT();
   
   // set everything to 0
   Servo_Duty(SERVO_CH_THROTTLE, SERVO_IN_MIN);
   Servo_Duty(SERVO_CH_AILERON, 0);
   Servo_Duty(SERVO_CH_ELEVATOR, 0);
   Servo_Duty(SERVO_CH_RUDDER, 0);
   
   SCI_OutString(1, "Throttle: 1->0    Q->1/2   A->1\r\n");
   SCI_OutString(1, "Aileron:  2->-1   W->0     S->1\r\n");
   SCI_OutString(1, "Elevator: 3->-1   E->0     D->1\r\n");
   SCI_OutString(1, "Rudder:   4->-1   R->0     F->1\r\n");
   
   for (;;)
   {			      
      c = SCI_InChar(1);
      
      switch (c) 
      {
         case '1':
            Servo_Duty(SERVO_CH_THROTTLE, SERVO_IN_MIN);
            break;
         case 'q':
            Servo_Duty(SERVO_CH_THROTTLE, 0);
            break;
         case 'a':
            Servo_Duty(SERVO_CH_THROTTLE, SERVO_IN_MAX);
            break;
         
         case '2':
            Servo_Duty(SERVO_CH_AILERON, SERVO_IN_MIN);
            break;
         case 'w':
            Servo_Duty(SERVO_CH_AILERON, 0);
            break;
         case 's':
            Servo_Duty(SERVO_CH_AILERON, SERVO_IN_MAX);
            break;
         
         case '3':
            Servo_Duty(SERVO_CH_ELEVATOR, SERVO_IN_MIN);
            break;
         case 'e':
            Servo_Duty(SERVO_CH_ELEVATOR, 0);
            break;
         case 'd':
            Servo_Duty(SERVO_CH_ELEVATOR, SERVO_IN_MAX);
            break;
         
         case '4':
            Servo_Duty(SERVO_CH_RUDDER, SERVO_IN_MIN);
            break;
         case 'r':
            Servo_Duty(SERVO_CH_RUDDER, 0);
            break;
         case 'f':
            Servo_Duty(SERVO_CH_RUDDER, SERVO_IN_MAX);
            break;                     
            
         default:
            break;
      }
   }   
}

// play with MicroSD
void PSSC5(void)  
{
   unsigned short i;
   long latitude, longitude, count;
   
    LEDS |= LED_RED;	  
   i = SD_Card_Init();
   if (i==0) LEDS &= ~LED_RED;
   
   // read the header
   LEDS |= LED_RED;
   i = SD_Read_Block(SD_Buffer, BLOCK_GPS_TRACK_IN);
   if (i==0) LEDS &= ~LED_RED;
   
   count = *(unsigned long *)SD_Buffer;
   
   for (i=0; i<count; i++)
   {
      latitude = ((unsigned long *)SD_Buffer)[i * 2 + 1];
      longitude = ((unsigned long *)SD_Buffer)[i * 2 + 2];
      SCI_OutChar(1, '('); SCI_OutLInt(1, latitude / 6L);  SCI_OutString(1, ", ");
      SCI_OutLInt(1, longitude / 6L);  SCI_OutString(1, ")\r\n");
   }
     
}


// Keep updating the GPS values, fire camera based on radio transmitter switches
void TestGPSmonitor(void) 
{    
   unsigned char prev_on = 1, cur_on = 0;
   unsigned char prev_shutter = 1, cur_shutter = 0;
      
   for (;;)
   {	
      // new gps data ready
      if (GPS_ReadyFlag()) 
      {        
         // translate the NMEA sentences to binary
         LEDS |= LED_YELLOW;
         GPS_Translate();         
         LEDS &= ~LED_YELLOW;        
      }     
      
      
      // get current switch state
      if (Radio_PulseWidth(0) < 35940) 
      {
         cur_on = 1;
      }
      else 
      {
         cur_on = 0;
      }
      
      if (Radio_PulseWidth(1) < 35940)
      {
         cur_shutter = 1;
      }
      else
      {
          cur_shutter = 0;
      }
      
      
      // act on switch changes
      if (cur_on == 1 && prev_on == 0) 
      {
         SCI_OutString(1, "Camera On\r\n");
         Camera_Send(CAMERA_COMMAND_ON);
      }
      else if (cur_on == 0 && prev_on == 1) 
      {
         SCI_OutString(1, "Camera Off\r\n");
         Camera_Send(CAMERA_COMMAND_OFF);
      }
      else if (cur_shutter == 1 && prev_shutter == 0)
      {
         SCI_OutString(1, "Shutter\r\n");
         Camera_Send(CAMERA_COMMAND_SHUTTER);
      }        
      
      prev_on = cur_on;
      prev_shutter = cur_shutter;
      
      Camera_HandleEvents();
   }   
}



// display a menu to test the functionality of various peripherals
void TestMenu(void)
{
   unsigned char c;
    
   // choose a test program to run
   for (;;)
   {
      SCI_OutString(1, "Choose a test:\r\n1. GPS\r\n2. SD\r\n3. Tracking\r\n4. Radio\r\n5. Camera\r\n6. Thermopiles\r\n");
      c = SCI_InChar(1);
      
      switch (c) 
      {
         case '1':    
            SCI_OutString(1, "\r\nGPS\r\n");        
            TestGPS();
            break;
         case '2':
            SCI_OutString(1, "\r\nSD\r\n");
            TestSD();
            break;
         case '3':
            SCI_OutString(1, "\r\nTracking\r\n");
            TestTracking();
            break;
         case '4':
            SCI_OutString(1, "\r\nRadio\r\n");
            TestRadio();
            break;
         case '5':
            SCI_OutString(1, "\r\nCamera\r\n");
            TestCamera();
            break;
         case '6':
            SCI_OutString(1, "\r\nThermopiles\r\n");
            TestThermopiles();
            break; 
         default:
            break;
      }
   }
}

/******************************* End of main.c ********************************/
