/********************************* nav.c ***************************************
*                                                                              *
*  Author: David Collins                                                       *
*  Project: Global Pigeon - ECE 477 Team 10, Spring 2008                       *
*                                                                              *
*  This file contains functions to control navigation and stabilization.       *
*  Navigation requires the ability to determine the angle and distance between *
*  two GPS waypoints.  Everything needed to be done using integers so a cosine *
*  lookup table was included to simplify operation.  A generic PID update      *
*  function is included which would be used for both navigation and            *
*  stabilization.  The RTI ISR is used to periodically measure the thermopile  *
*  values and then store various inputs onto the SD card.  Autopilot has not   *
*  been implemented yet due to the problems with radio interference.           *
*                                                                              *
*******************************************************************************/


#include <mc9s12a64.h>
#include "nav.h"
#include "pigeon.h"
#include "adc.h"
#include "sci.h"
#include "sd.h"
#include "radio.h"



// internal functions

// sends commands to servos at regular intervals
void Nav_Handler(void);

// print out the averaged thermopile values
void PrintThermopiles(void);

// scale a thermopile value so that they all have the same base value
unsigned short Therm_Scaled(unsigned char channel);

// calculate the square root of a long input value
// source: http://www.codecodex.com/wiki/index.php?title=Calculate_an_integer_square_root
unsigned long lsqrt(unsigned long y);



// performs thermopile scaling and display
void Test_ThermScale(void);

// Store GPS data as well as thermopile readings and radio signals to the SD card
void TestLogging(void);

// set up initial used in the data logging
void RTI_Init(void);


// cos(ang) * 256 with ang in degree
// used for scaling down longitude values based on latitude
// cosAdjust[0] corresponds to 4 deg as 0-3deg == 256
const unsigned char cosAdjust[87] = {
   255, 255, 255, 254, 254, 253, 252, 251, 250, 249, 248, 247, 246, 245, 243,
   242, 241, 239, 237, 236, 234, 232, 230, 228, 226, 224, 222, 219, 217, 215,
   212, 210, 207, 204, 202, 199, 196, 193, 190, 187, 184, 181, 178, 175, 171,
   168, 165, 161, 158, 154, 150, 147, 143, 139, 136, 132, 128, 124, 120, 116,
   112, 108, 104, 100,  96,  92,  88,  83,  79,  75,  71,  66,  62,  58,  53,
    49,  44,  40,  36,  31,  27,  22,  18,  13,   9,   4,   0};	  



// determine how many points will fit in a block; is the max index
#define STATE_BLOCK_MAX ((SD_BLOCK_SIZE / sizeof(STATE_DATA)))



// amount to downshift
#define THERM_SCALE_SHIFT 10

long therm_scale_mul[THERM_COUNT] = {1024, 1029, 992, 1054};
long therm_scale_add[THERM_COUNT] = {0, -3, -6, -21};
 


// counter for rti interrupt
unsigned int rti_count = 0;

// holds the thermopile sums
unsigned int therm_sum[THERM_COUNT] = {0, 0, 0, 0};
unsigned int therm_value[THERM_COUNT] = {0, 0, 0, 0};

unsigned int vsense_sum = 0, vsense_value = 0;

PID therm3pid;
long int therm3set;

extern unsigned char SD_Buffer[SD_BLOCK_SIZE];

unsigned char RTI_DataReady;

 	  
// number of blocks used on the SD card 
unsigned long LogCount;

// number of entries logged in current block
int LogBufCount;



// set up registers needed for navigation
void Nav_Init(void)
{

   // set up initial used in the data logging
   RTI_Init();


   // set up RTI to poll thermopile values and round them
   
   // set RTI to interrupt at 325.52 Hz = 8 MHz OSCCLK / 3*2^13
   // this was chosen so that 10 samples averaged before a 10 Hz servo adjustment
   // would correspond to the last 1/3 of the time between actions
   RTICTL = 0x42;
   
   // enable RTI
   CRGINT = 0x80;
   
   therm3set = 440L; 
   
   therm3pid.Kp = 10;
   therm3pid.Kd = 10;
   therm3pid.Ki = 1;
   therm3pid.sumError = 0;
   therm3pid.prevError = 0;
   
   therm3pid.outputMin = -1500;
   therm3pid.outputMax = 1500;
   therm3pid.sumErrorMin = -100;
   therm3pid.sumErrorMax = 100;
   
   RTI_DataReady = 0;    
}


interrupt 7 void RTI_ISR(void)
{
   int i;
   
   // clear the RTIF flag
   CRGFLG = 0x80;
   
   if (rti_count < NAV_AVERAGE_COUNT)
   {
      // add values to running sum      
      therm_sum[0] += ADC_In(ADC_CH_THERM0);
      therm_sum[1] += ADC_In(ADC_CH_THERM1);
      therm_sum[2] += ADC_In(ADC_CH_THERM2);
      therm_sum[3] += ADC_In(ADC_CH_THERM3);
      
      vsense_sum += ADC_In(ADC_CH_VSENSE);     
   }
   else if (rti_count == NAV_AVERAGE_COUNT)
   {
   
      // calculate averaged values
      for (i=0; i<THERM_COUNT; i++)
      {
         therm_value[i] = therm_sum[i] >> NAV_SHIFT;
         therm_sum[i] = 0;
      }
      
      vsense_value = vsense_sum >> NAV_SHIFT;
      vsense_sum = 0;
    
      // perform all timely events  
      Nav_Handler();
   }
   else if (rti_count >= NAV_STABILIZE_COUNT)
   {
      // wrap back around
      rti_count = 0;
   }
   
   rti_count++; 
}


// sends commands to servos at regular intervals
void Nav_Handler(void)
{  
   // Store GPS data as well as thermopile readings and radio signals to the SD card
   TestLogging();
   
   RTI_DataReady = 1;

   // performs thermopile scaling and display
   // Test_ThermScale();   
}


// print out the averaged thermopile values
void PrintThermopiles(void)
{
   SCI_OutChar(1, '\f');
   SCI_OutString(1, "Vsense: "); SCI_OutUInt(1, vsense_value); SCI_OutString(1, "\r\n");
   
   // Raw
   SCI_OutString(1, "Raw\r\n");
   SCI_OutString(1, "Therm0: "); SCI_OutUInt(1, therm_value[0]); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "Therm1: "); SCI_OutUInt(1, therm_value[1]); SCI_OutString(1, "\r\n");  
   SCI_OutString(1, "Therm2: "); SCI_OutUInt(1, therm_value[2]); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "Therm3: "); SCI_OutUInt(1, therm_value[3]); SCI_OutString(1, "\r\n");	     
   
   // Scaled 
   SCI_OutString(1, "Scaled\r\n");
   SCI_OutString(1, "Therm0: "); SCI_OutULInt(1, Therm_Scaled(0)); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "Therm1: "); SCI_OutULInt(1, Therm_Scaled(1)); SCI_OutString(1, "\r\n");  
   SCI_OutString(1, "Therm2: "); SCI_OutULInt(1, Therm_Scaled(2)); SCI_OutString(1, "\r\n");
   SCI_OutString(1, "Therm3: "); SCI_OutULInt(1, Therm_Scaled(3)); SCI_OutString(1, "\r\n");    
}



// outputs the new control output value
long int Nav_PID(long int error, PID *pid)
{
   long int P, I, D, out;
   
   // Proportional   
   P = pid->Kp * error;
   
   // Integral
   pid->sumError += error;
   if (pid->sumError > pid->sumErrorMax)
   {
      pid->sumError = pid->sumErrorMax;
   }
   else if(pid->sumError < pid->sumErrorMin)
   {
      pid->sumError = pid->sumErrorMin;
   }
      
   I = pid->Ki * pid->sumError;
   
   // Derivative
   D = pid->Kd * (error - pid->prevError);
   pid->prevError = error;
   
   
   out = P + I + D;   
   if (out > pid->outputMax)
   {
      out = pid->outputMax;
   }
   else if (out < pid->outputMin)
   {
      out = pid->outputMin;
   }
   
   
   return out;
}

// scale a thermopile value so that they all have the same base value
unsigned short Therm_Scaled(unsigned char channel)
{
   long out;
   out = therm_value[channel];
   out *= therm_scale_mul[channel];
   out >>= THERM_SCALE_SHIFT;
   out += therm_scale_add[channel];
   
   return (unsigned short)out;
}






// calculates the distance (ft) and bearing (deg) between two waypoints
// bearing vector goes from p1 to p2
void Nav_CalcDifference(unsigned long *pDist, unsigned long *pBearing, const GPS_DATA *p1, const GPS_DATA *p2) {
   long dlat, dlong, temp;
   int downshift=0;
   unsigned char target, type, low=0, mid=43, high=86;
   unsigned long udlat, udlong, utemp, utemp2, uval;

   //printf("p1: (%ld, %ld), p2: (%ld, %ld)\n", p1->latitude, p1->longitude, p2->latitude, p2->longitude);
   // calculate the differences between the waypoints
   // differences currently in units of minutes of angle
   dlat = p2->latitude - p1->latitude;
   dlong = p2->longitude - p1->longitude;

   // convert from minute to degree, rounding to the nearest degree
   if (p1->latitude > 0) temp = p1->latitude / 600000L;
   else temp = p1->latitude / -600000L;

   // scale dlong as longitude degrees get closer together at higher latitudes
   // no change for latitudes <4deg
   if (temp > 90) {
      dlong = 0;      
   }
   else if (temp > 3) {
      dlong *= cosAdjust[temp - 4];
      dlong >>= 8;
   }

   // absolute values
   udlat = (dlat > 0 ? dlat : -dlat);
   udlong = (dlong > 0 ? dlong : -dlong);

   //printf("dlat:  %ld\ndlong: %ld\n", dlat, dlong);

   // downshift the inputs so that the square root doesn't fail
   while (udlat > 0x8000 || udlong > 0x8000) {
      udlat >>= 1;
      udlong >>= 1;
      downshift++;
   }

   //printf("Downshift=%d, udlat=%ld, udlong=%ld\n", downshift, udlat, udlong);

   // calculate distance
   *pDist = udlat * udlat + udlong * udlong;
   *pDist = lsqrt(*pDist);

   // scale back up
   // pDist is units of minutes * 10000
   *pDist <<= downshift;

   //printf("Distance: %ld (min)\n", *pDist);
   


   // absolute values
   udlat = (dlat > 0 ? dlat : -dlat);
   udlong = (dlong > 0 ? dlong : -dlong);

   utemp = *pDist;
   downshift = 0;

   // pick the triangle leg such that sine or cosine is small
   if (udlat < udlong) { 
      uval = udlat;
      type = 0;
   }
   else {
      uval = udlong;
      type = 1;
   }

   //printf("uval: %ld\n", uval);;

   // ensure it doesn't go out of range
   while (uval > 0x00FFFFFFuL) {
      uval >>= 1;
      utemp >>= 1;
      downshift++;
   }

   // calulate the target sine or cosine
   utemp2 = uval << 8; 
   //printf("uval: %ld  utemp2: %ld\n", uval, utemp2);
   utemp2 /= utemp;
   //printf("utemp: %ld  utemp2: %ld\n", utemp, utemp2);
   target = (unsigned char)utemp2;
   //printf("Target: %d\n", target);

   // convert from minute * 10000 to feet
   *pDist = (utemp * 155) >> (8 - downshift);
   //printf("Distance: %ld (ft)\n", *pDist);


   //printf("Type: %d\n", type);
   if (uval == 0) {
      // corner case of vertical or horizontal travel
      mid = 90;
   }
   else {

      // find the angle using bisection   
      while (low < mid && mid < high) {
         //printf("L: %2d, M: %2d, H: %2d\n", low, mid, high);
         if (target == cosAdjust[mid]) {
            break;
         }
         else if (target < cosAdjust[mid]) {
            low = mid;
         }
         else {
            high = mid;
         }
         mid = (high + low) >> 1;      
      }

      // make mid an angle away from north
      mid += 4;
   }

   if (type) mid = 90 - mid;

   //printf("Angle: %d\n", mid);

   if (dlat >= 0 && dlong >= 0) {
      // quandrant I
      *pBearing = mid;
   }
   else if (dlat >= 0 && dlong <= 0) {
      // quandrant II
      *pBearing = 360 - mid;
   }
   else if (dlat <= 0 && dlong <= 0) {
      // quandrant III
      *pBearing = mid + 180;
   }
   else {
      // quandrant IV
      *pBearing = 180 - mid;
   }

}



// calculate the square root of a long input value
// source: http://www.codecodex.com/wiki/index.php?title=Calculate_an_integer_square_root
unsigned long lsqrt(unsigned long y)
{
   unsigned long x_old, x_new;
   unsigned long testy;
   int nbits;
   int i;

   if (y == 0) {
      return 0L;
   }

   // select a good starting value using binary logarithms: 
   nbits = (sizeof(y) * 8);
   for (i = 4, testy = 16L;; i += 2, testy <<= 2L) {
      if (i >= nbits || y <= testy) {
         x_old = (1L << (i >> 1));
         // x_old = closest power of 2 to sqrt(testy) 
         break;
      }
   }
   
   // x_old >= sqrt(y)
   // use the Babylonian method to arrive at the integer square root:
   // i should never be very big, but set a cap just in case

   for (i=0;i<50;i++) { 
      x_new = (y / x_old + x_old) >> 1;
      if (x_old <= x_new)
            break;
      x_old = x_new;
   }   
   
   // make sure that the answer is right: 
   /*
   if (x_old * x_old > y || (x_old + 1) * (x_old + 1) <= y) {
      //printf("Error in lsqrt(), bad output\n");
      return 0L;
   }
   */
   
   return x_old;
}


// performs thermopile scaling and display
void Test_ThermScale(void)
{
   long int output;
   
   // alternate green LED
   LEDS ^= LED_GREEN;
   
   
   // output the thermopile values for debugging
   PrintThermopiles();  
   
   output = Nav_PID(therm_value[3] - therm3set, &therm3pid);
   SCI_OutString(1, "Output: "); SCI_OutLInt(1, output); SCI_OutString(1, "\r\n"); 
}




// set up initial used in the data logging
void RTI_Init(void)
{
   unsigned short i;
   
   LogCount = 0;
   LogBufCount = 0;
   
   
   LEDS |= LED_RED;	  
   i = SD_Card_Init();
   if (i==0) LEDS &= ~LED_RED;  
   
   
   for (i=0; i<SD_BLOCK_SIZE; i++) SD_Buffer[i] = 0;
   // clear the header
   // i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   
   // read the header
   LEDS |= LED_RED;
   i = SD_Read_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
   if (i==0) LEDS &= ~LED_RED;
   
   // copy out the first 4 bytes as an integer
   LogCount = *(unsigned long *)SD_Buffer;
      
}

  

// Store GPS data as well as thermopile readings and radio signals to the SD card
void TestLogging(void) 
{	    
   unsigned short i;
   STATE_DATA *state;
     
   // alternate green LED
   LEDS ^= LED_GREEN;
   
   // get the storage location
   state = &((STATE_DATA *)SD_Buffer)[LogBufCount]; 
                          
   // copy the waypoint data into the SD buffer
   GPS_Value(&(state->gps));
   
   // copy in radio values
   for (i=0; i<6; i++)
   {
      state->radio[i] = Radio_PulseWidth(i);
   }
   
   // copy in thermopile values
   for (i=0; i<4; i++)
   {
      state->therm[i] = therm_value[i];
   }
                 
      
   LogBufCount++;
   
   if (LogBufCount >= STATE_BLOCK_MAX)
   {
      LogBufCount = 0;
      
      // write the data to the SD card
      LogCount++;
      LEDS |= LED_RED;
      i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT + LogCount);
      if (i==0) LEDS &= ~LED_RED;
      
      // write the count back to the header
      *(unsigned long *)SD_Buffer = LogCount;
      LEDS |= LED_RED;
      i = SD_Write_Block(SD_Buffer, BLOCK_GPS_TRACK_OUT);
      if (i==0) LEDS &= ~LED_RED;
   }	 
   
}

/******************************** End of nav.c ********************************/
