/******************************* radio.c ***************************************
*                                                                              *
*  Author: David Collins                                                       *
*  Project: Global Pigeon - ECE 477 Team 10, Spring 2008                       *
*                                                                              *
*  Uses the Timer module to receive servo pulses using 6 input capture pins.   *
*  The pules are 1ms to 2ms in duration every 20ms.                            *
*                                                                              *
*  This driver is based upon measuring the number of clock cycles between a    *
*  rising and falling edge.  However, the input capture registers are only     *
*  16bit and the free running clock needs to increment at every clock cycles.  *
*  This means that with a 24MHz clock, the maximum pulse width that can be     *
*  measured is 2.73ms.  Exceeding this value will cause the pulse width        *
*  variables to wrap back around to 0.                                         *
*                                                                              *
*******************************************************************************/


#include <hidef.h> 
#include <mc9s12a64.h>

#include "radio.h"
#include "timer.h"
#include "servo.h"

/* 
   9S12A64 Pins
   PTT0 - PTT5
*/
   

// used for timing the incoming servo control pulses
unsigned short radio_start[RADIO_MAX_CHANNELS];
unsigned short radio_stop[RADIO_MAX_CHANNELS];
unsigned short radio_pulse[RADIO_MAX_CHANNELS];

// high for each channel to indicate new data
unsigned char radio_flag;

// determines which radio channels pass directly to PWM on interrupt
unsigned char radio_passthru;

// configures the timer input capture to allow for receiving servo signals
void Radio_Init(void)
{					
   int i;
   
   // zero the counters
   for (i=0; i<RADIO_MAX_CHANNELS; i++)
   {
      radio_start[i]=0;
      radio_stop[i]=0;
      radio_pulse[i]=0;
   }
   
   radio_passthru = 0xFF;
   
   
   // set PTT0-PTT5 for input capture
   TIOS &= ~0b00111111;
   
   // input capture on rising edge
   TCTL3 = 0b00000101; 
   TCTL4 = 0b01010101;
   
   // set PTT0-PTT5 as input
   DDRT &= ~0b00111111;
   
   // enable a pull up
   //PERT |= 0b00111111;
   //PPST = 0;
   
   // enable interrupts on IOC0-IOC5
   TIE |= 0b00111111;

}


// Chose to duplicate ISR code so that ISRs run as fast as possible


// interrupt handler for IOC0 channel
// Radio channel Aux
interrupt 8 void IOC0_Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x01;
   
   if (TCTL4 & 0b00000001) // rising edge
   {
      // switch to falling edge
      TCTL4 &= 0b11111100;
      TCTL4 |= 0b00000010;
      
      radio_start[0] = TC0;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL4 &= 0b11111100;
      TCTL4 |= 0b00000001;
            
      radio_stop[0] = TC0;      
      radio_pulse[0] = radio_stop[0] - radio_start[0];
      radio_flag |= 0x01;     
   }
}


// interrupt handler for IOC1 channel
// Radio channel Auto
interrupt 9 void IOC1_Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x02;
   
   if (TCTL4 & 0b00000100) // rising edge
   {
      // switch to falling edge
      TCTL4 &= 0b11110011;
      TCTL4 |= 0b00001000;
      
      radio_start[1] = TC1;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL4 &= 0b11110011;
      TCTL4 |= 0b00000100;
            
      radio_stop[1] = TC1;      
      radio_pulse[1] = radio_stop[1] - radio_start[1];
      radio_flag |= 0x02;     
   }
}


// interrupt handler for IOC2 channel
// Radio channel Elevator
interrupt 10 void IOC2Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x04;
   
   if (TCTL4 & 0b00010000) // rising edge
   {
      // switch to falling edge
      TCTL4 &= 0b11001111;
      TCTL4 |= 0b00100000;
      
      radio_start[2] = TC2;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL4 &= 0b11001111;
      TCTL4 |= 0b00010000;
            
      radio_stop[2] = TC2;      
      radio_pulse[2] = radio_stop[2] - radio_start[2];
      radio_flag |= 0x04;     
      
      if (radio_passthru & RADIO_MASK_ELEVATOR) 
      {
         PWM_Duty16(SERVO_CH_ELEVATOR, radio_pulse[RADIO_CH_ELEVATOR] >> 3);
      }
   }
}


// interrupt handler for IOC3 channel
// Radio channel Rudder
interrupt 11 void IOC3_Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x08;
   
   if (TCTL4 & 0b01000000) // rising edge
   {
      // switch to falling edge
      TCTL4 &= 0b00111111;
      TCTL4 |= 0b10000000;
      
      radio_start[3] = TC3;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL4 &= 0b00111111;
      TCTL4 |= 0b01000000;
            
      radio_stop[3] = TC3;      
      radio_pulse[3] = radio_stop[3] - radio_start[3];
      radio_flag |= 0x08;  
      
      if (radio_passthru & RADIO_MASK_RUDDER) 
      {
         PWM_Duty16(SERVO_CH_RUDDER, radio_pulse[RADIO_CH_RUDDER] >> 3);
      }   
   }
}


// interrupt handler for IOC4 channel
// Radio channel Aileron
interrupt 12 void IOC4_Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x10;
   
   if (TCTL3 & 0b00000001) // rising edge
   {
      // switch to falling edge
      TCTL3 &= 0b11111100;
      TCTL3 |= 0b00000010;
      
      radio_start[4] = TC4;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL3 &= 0b11111100;
      TCTL3 |= 0b00000001;
            
      radio_stop[4] = TC4;      
      radio_pulse[4] = radio_stop[4] - radio_start[4];
      radio_flag |= 0x10;  
      
      if (radio_passthru & RADIO_MASK_AILERON) 
      {
         PWM_Duty16(SERVO_CH_AILERON, radio_pulse[RADIO_CH_AILERON] >> 3);
      }   
   }
}


// interrupt handler for IOC5 channel
// Radio channel Throttle
interrupt 13 void IOC5_Handler(void)  
{	    
   // acknowledge the interrupt
   TFLG1 = 0x20;
   
   if (TCTL3 & 0b00000100) // rising edge
   {
      // switch to falling edge
      TCTL3 &= 0b11110011;
      TCTL3 |= 0b00001000;
      
      radio_start[5] = TC5;         
   }
   else                   // falling edge
   {
      // switch to rising edge         
      TCTL3 &= 0b11110011;
      TCTL3 |= 0b00000100;
            
      radio_stop[5] = TC5;      
      radio_pulse[5] = radio_stop[5] - radio_start[5];
      radio_flag |= 0x20;  
      
      if (radio_passthru & RADIO_MASK_THROTTLE) 
      {
         PWM_Duty16(SERVO_CH_THROTTLE, radio_pulse[RADIO_CH_THROTTLE] >> 3);
      }   
   }
}




// returns the current pulse width measured on a radio channel
unsigned short Radio_PulseWidth(unsigned char channel)
{
   if (channel < RADIO_MAX_CHANNELS)
   {
      return radio_pulse[channel];
   }
   else
   {
      return -1;
   }
}


// returns true if the flag has be asserted, indicating a new recieved pulse
unsigned char Radio_ReadyFlag(unsigned char channel)
{
   if (channel < RADIO_MAX_CHANNELS)
   {        
      return (radio_flag & (1<<channel));
   }
   else
   {
      return 0;
   }
}


// acknowledges a flag by setting it low
void Radio_DisassertFlag(unsigned char channel)
{
   if (channel < RADIO_MAX_CHANNELS)
   {        
      radio_flag &= ~(1<<channel);
   }
}

// turns on radio pass through for channels specified in mask
void Radio_PassthruSet(unsigned char mask)
{
   radio_passthru |= mask;
}

// turns off radio pass through for channels specified in mask
void Radio_PassthruUnset(unsigned char mask)
{
   radio_passthru &= ~mask;
}

/******************************* End of radio.c *******************************/
