/******************************* servo.c ***************************************
*                                                                              *
*  Author: David Collins                                                       *
*  Project: Global Pigeon - ECE 477 Team 10, Spring 2008                       *
*                                                                              *
*  Uses the PWM module to send servo pulses using all port P pins.  However,   *
*  only pins 1,3,5, and 7 of port P have an output signal. The pulse are 1ms   *
*  to 2ms in duration every 20ms to indicate full left to full right.  The PWM *
*  channels operate in 16bit concatenated mode.                                *
*                                                                              *
*******************************************************************************/


#include <hidef.h> 
#include <mc9s12a64.h>

#include "servo.h"
#include "pigeon.h"

/* 
   9S12A64 Pins
   PTP0-PTP7
   
   Outputs on PTP1, PTP3, PTP5, PTP7 
*/


// used to interact with duty and period registers
unsigned short *PWM_Duty16_Offset = (unsigned short *)&PWMDTY0;
unsigned short *PWM_Period16_Offset = (unsigned short *)&PWMPER0;


// sets up the PWM registers for 20ms period and servo centered initial duty
void Servo_Init(void) 
{
   // Disable PWM module for PTT0,1; setting to 16bit mode works only when disabled   
   PWME = 0;
      
   // Duty On time is logic high   
   PWMPOL = 0xFF;
   
   // use clock A and B respectively
   PWMCLK = 0;
   
   // left aligned PWM
   PWMCAE = 0; 
   
   // clock A prescaler of 8
   // for a 16bit PWM and 24MHz clock, this leads to 60000 period = 20ms
   // thus duty values of 3000 to 6000 are approprate servo values 
   PWMPRCLK = 0b00110011;
   
   // 100Hz
   //PWMPRCLK = 0b00100010;
   
   // enable PWM channel concatenation to get 16bit PWM
   PWMCTL |= 0xF0;
  
   PWM_Period16(0, SERVO_PWM_PERIOD);
   
   Servo_Duty(SERVO_CH_THROTTLE, SERVO_IN_MIN);
   Servo_Duty(SERVO_CH_AILERON, 0);
   Servo_Duty(SERVO_CH_ELEVATOR, 0);
   Servo_Duty(SERVO_CH_RUDDER, 0);
   
   // Enable PWM module   
   PWME = 0xFF;
   
}


// Tested with a Hitec HS-322HD
// It was able to operate at 50Hz and 100Hz, but seized at 150Hz and 200Hz.
// The total range of inputs with period=255, is duty=15 to 62, this corresponds to about 200deg of rotation 



// sets the duty cycle of selected 16bit PWM channel
// only two 16bit channels are available
void PWM_Duty16(unsigned char channel, unsigned short value)
{
   if (channel < SERVO_COUNT)	 // ensure only valid channels can be set
   {        
      PWM_Duty16_Offset[channel] = value;
   }
}

// sets the period of selected 16bit PWM channel
void PWM_Period16(unsigned char channel, unsigned short value)
{
   if (channel < SERVO_COUNT)	 // ensure only valid channels can be set
   {        
      PWM_Period16_Offset[channel] = value;
   }
}

// sets the pulse width of a servo channel
// this corresponds to a 16bit PWM channel
// input values are SERVO_IN_MIN to SERVO_IN_MAX
void Servo_Duty(unsigned char channel, short value)
{
   if (channel < SERVO_COUNT)	 // ensure only valid channels can be set
   {  
      // saturate inputs to handle out of range values
      if (value > SERVO_IN_MAX) value = SERVO_IN_MAX;
      if (value < SERVO_IN_MIN) value = SERVO_IN_MIN; 
      
      // scale to get correct servo pwm duty cycle
      
      // this simple technique ensures that the zero point
      // is handle correctly, but could lead to high or low clipping
      // a multiplicative scaling method would be more appropriate
      // but more difficult to do with integers
      value += SERVO_OUT_PWM_ZERO;      
      if (value > SERVO_OUT_PWM_MAX) value = SERVO_OUT_PWM_MAX;
      if (value < SERVO_OUT_PWM_MIN) value = SERVO_OUT_PWM_MIN;
           
      PWM_Duty16_Offset[channel] = value;
   }
}

/******************************* End of servo.c *******************************/
