/******************************* gps.c *******************************/

// Communicates with an ETek EB-85A GPS module
// However, the usage of standard NMEA sentences should allow this source
// code to work on other GPS modules.
 
//#include <mc9s12c32.h>
#include "gps.h"
//#include "constants.h"


#include <stdio.h> 
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <netinet/in.h>

const char NMEA_SET_MIN[] = "$PMTK314,0,1,0,1,5,0,0,0,0,0,0,0,0,0,0,0,0*2D\r\n";
//const char NMEA_SET_MIN[] = "$PMTK314,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
//const char NMEA_SET_MIN[] = "$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
const char NMEA_SET_WAAS[] = "$PMTK313,1*2E\r\n$PMTK301,2*2E\r\n";


#define GPS_NUM_ACCEPTED 3

const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPRMC", "GPGGA", "GPGSA"};
//const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPVTG", "GPGGA"};
//const char gpsNmeaAccepted[GPS_NUM_ACCEPTED][6] = {"GPRMC", "GPGGA"};


#define GPS_NUM_BUFFERS 5
#define GPS_BUFFER_SIZE 255

// head and tail points for buffer selection
int gpsHead, gpsTail;

// GPS input buffer
char gpsBuffer[GPS_NUM_BUFFERS][GPS_BUFFER_SIZE];

// current writing position in each GPS buffer
int gpsBufferPos[GPS_NUM_BUFFERS];

// flag is true while currently receiving a NMEA sentence 
char gpsInSentence;
// flag true if NMEA sentence
char gpsInAccepted;



// bits for accepted sentences set high when received
int gpsFlag;


GPS_DATA gpsData;

// initial the GPS module and the UART port connecting to it
int GPS_Init(void) {
   int fd;
   struct termios options;
   int i;
   unsigned char *p;

   // initialize gps buffers to 0
   gpsHead = gpsTail = 0;
   for (i=0; i<GPS_NUM_BUFFERS; i++) {
      gpsBufferPos[i] = 0;
   }

   gpsInSentence = 0;
   gpsInAccepted = 0;
   gpsFlag = 0;

   // zero out the gps data structure
   p = (unsigned char *)&gpsData;
   for (i=0; i<sizeof(GPS_DATA); i++) {
      p[i] = 0;
   }

   // open COM Port 1
   fd = open("/dev/ttyS0", O_RDWR | O_NOCTTY | O_NDELAY);

   if (fd == -1) {
      printf("Could not open COM Port 1 (/dev/ttyS0)\n");
      return -1;
   }
   
   // non-blocking reads
	fcntl(fd, F_SETFL, FNDELAY);

  
   // Get the current options for the port
   tcgetattr(fd, &options);
    
   // Set the baud rates to 38400
   cfsetispeed(&options, B38400);
   cfsetospeed(&options, B38400);

   // Enable the receiver and set local mode
   options.c_cflag |= (CLOCAL | CREAD);

   // Set for 8 bit, no parity
   options.c_cflag &= ~PARENB;
   options.c_cflag &= ~CSTOPB;
   options.c_cflag &= ~CSIZE;
   options.c_cflag |= CS8;

   // Set for no flow control
   options.c_cflag &= ~CRTSCTS;
   options.c_iflag &= ~(IXON | IXOFF | IXANY);

   // Set the new options for the port
   tcsetattr(fd, TCSANOW, &options);

   // send NMEA command to set for minimal sentence output
   write(fd, NMEA_SET_MIN, strlen(NMEA_SET_MIN));

   // send NMEA command to enable WAAS operation
   write(fd, NMEA_SET_WAAS, strlen(NMEA_SET_WAAS));


   return (fd);
}


// handle incoming data
void GPS_Receive(int fd) {
   int n, i;
   char c;
   
   n = read(fd, &c, 1);
   if (n > 0) {
      if (!gpsInSentence) {
         if (c == '$') {            
            gpsInSentence = 1;
         }
      }
      else if (!gpsInAccepted) {
         if (c == ',') {
            // TalkerID finished, check if it is accepted            
            for (i=0; i<GPS_NUM_ACCEPTED; i++) {               
               if (strncmp(gpsBuffer[gpsTail], gpsNmeaAccepted[i], gpsBufferPos[gpsTail]) == 0) {
                  // it is an accepted sentence
                  break;
               }
            }
            if (i<GPS_NUM_ACCEPTED) {               
               // replace first character in buffer with accepted NMEA index
               gpsBuffer[gpsTail][0] = '0' + i;
               gpsBuffer[gpsTail][1] = ',';
               gpsBufferPos[gpsTail] = 2;
               gpsInAccepted = 1;
            }
            else {               
               // this is not an accepted sentence, so set for not in sentence
               gpsInSentence = 0;
               // move position back to start of buffer
               gpsBufferPos[gpsTail] = 0;
            }
         }
         else {
            // add character to buffer
            gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = c;
            gpsBufferPos[gpsTail]++;
         }
      }
      else {
         // is in an accepted NMEA sentence, copy into buffer if not end
                  
         if (c == '*') {            
            // end of the sentence
            gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = '\0';
            gpsBufferPos[gpsTail]++;

            // set for not in sentence and not accepted
            gpsInSentence = 0;
            gpsInAccepted = 0;

            // move to next buffer with circular add
            gpsTail++;
            if (gpsTail == GPS_NUM_BUFFERS) gpsTail=0;            

            // move to the start of the buffer
            gpsBufferPos[gpsTail] = 0;
         }
         else {
            // just copy the char
            gpsBuffer[gpsTail][gpsBufferPos[gpsTail]] = c;
            gpsBufferPos[gpsTail]++;
         }
      }
   }
}


// non-interrupt function for translating NMEA sentences into useful data
void GPS_Translate(void) {
   int i, j, neg, max;
   char *buf;

   
   if (gpsTail != gpsHead) {
      // new data is ready in the buffer
      //printf("%s\n", gpsBuffer[gpsHead]);

      buf = gpsBuffer[gpsHead];
      max = gpsBufferPos[gpsHead];
      
      i = buf[0] - '0';
      if (i == 0) {
         // RMC

         // get time
         gpsData.hr = (buf[2] - '0')*10 + (buf[3] - '0');
         gpsData.min = (buf[4] - '0')*10 + (buf[5] - '0');
         gpsData.sec = ((((buf[6]-'0')*10+(buf[7]-'0'))*10+(buf[9]-'0'))*10+(buf[10]-'0'))*10+(buf[11]-'0');
         
         gpsData.latitude = ((buf[15]-'0')*10+(buf[16]-'0'))*6;         
         // read in latitude
         for (i=17; i<24; i++) {
            if (buf[i] >= '0' && buf[i] <= '9') {
               if (i>17) gpsData.latitude *= 10;
               gpsData.latitude += buf[i] - '0';               
            }
         }
         if (buf[25] == 'S') {
            // make negative for southern hemisphere
            gpsData.latitude = 0 - gpsData.latitude;            
         }

         gpsData.longitude = (((buf[27]-'0')*10+(buf[28]-'0'))*10+(buf[29]-'0'))*6;
         // read in latitude
         for (i=30; i<37; i++) {
            if (buf[i] >= '0' && buf[i] <= '9') {
               if (i>30) gpsData.longitude *= 10;
               gpsData.longitude += buf[i] - '0';
            }
         }
         if (buf[38] == 'W') {
            // make negative for west
            gpsData.longitude = 0 - gpsData.longitude;
         }

         // get speed
         gpsData.speed = 0;
         i=40;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.speed *= 10;
               gpsData.speed += buf[i] - '0';
            }
            i++;
         }

         // get bearing
         gpsData.bearing = 0;
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.bearing *= 10;
               gpsData.bearing += buf[i] - '0';
            }
            i++;
         }

         // get date
         i++;
         gpsData.month = (buf[i]-'0')*10+(buf[i+1]-'0');
         i+=2;
         gpsData.day = (buf[i]-'0')*10+(buf[i+1]-'0');
         i+=2;
         gpsData.year = (buf[i]-'0')*10+(buf[i+1]-'0');
         
         //
         i+=5;
         switch (buf[i]) {
            case 'N':
               gpsData.navMode = 0;
               break;
            case 'A':
               gpsData.navMode = 1;
               break;
            case 'D':
               gpsData.navMode = 2;
               break;
            case 'E':
               gpsData.navMode = 3;
               break;
         }
         
         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPRMC;
         
      }
      else if (i == 1) {
         // GGA

         // get number of satellites
         gpsData.numSV = 0;
         i=40;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.numSV *= 10;
               gpsData.numSV += buf[i] - '0';
            }
            i++;
         }


         // altitude
         gpsData.altitude = 0;
         i++;
         j=0;
         while (j<1) {
            if (buf[i] == ',') j++;
            i++;
         }

         neg = 0;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.altitude *= 10;
               gpsData.altitude += buf[i] - '0';
            }
            else if (buf[i] == '-') {
               neg = 1;
            }
            i++;
         }
         // negate if needed
         if (neg) gpsData.altitude = 0 - gpsData.altitude;



         // get geoidal separation
         gpsData.geoidSep = 0;
         i+=3;         
         neg = 0;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.geoidSep *= 10;
               gpsData.geoidSep += buf[i] - '0';
            }
            else if (buf[i] == '-') {
               neg = 1;
            }
            i++;
         }
         // negate if needed
         if (neg) gpsData.geoidSep = 0 - gpsData.geoidSep;

         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPGGA;
         
      }
      else if (i == 2) {
         // GSA

         // get fix mode
         gpsData.fix = buf[4] - '0';
         if (gpsData.fix == 1) gpsData.fix=0;

         // move to the DOPs
         i=6;
         j=0;
         while (j<12) {
            if (buf[i] == ',' && i < max) j++;
            i++;
         }

         // get PDOP
         gpsData.PDOP = 0;         
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.PDOP *= 10;
               gpsData.PDOP += buf[i] - '0';
            }
            i++;
         }

         // get PDOP
         gpsData.HDOP = 0; 
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.HDOP *= 10;
               gpsData.HDOP += buf[i] - '0';
            }
            i++;
         }

         // get PDOP
         gpsData.VDOP = 0; 
         i++;
         while (buf[i] != ',' && i < max) {         
            if (buf[i] >= '0' && buf[i] <= '9') {
               gpsData.VDOP *= 10;
               gpsData.VDOP += buf[i] - '0';
            }
            i++;
         }

         // raise the ready flag
         gpsFlag |= GPS_NMEA_GPGSA;         

      }



      // move to next buffer with circular add
      gpsHead++;
      if (gpsHead == GPS_NUM_BUFFERS) gpsHead=0;
   }


   // print everything when both are new
   if ((gpsFlag & (GPS_NMEA_GPRMC | GPS_NMEA_GPGGA)) == (GPS_NMEA_GPRMC | GPS_NMEA_GPGGA)) {
      gpsFlag &= ~(GPS_NMEA_GPRMC | GPS_NMEA_GPGGA);

      OutputGPS(&gpsData);
   }
}

// converts the endianness of incoming data from the microcontroller
// The micro uses big endian, but linux uses little endian
void ConvertGPS(GPS_DATA *data) {
   data->latitude = ntohl(data->latitude);
   data->longitude = ntohl(data->longitude);
   data->altitude = ntohl(data->altitude);
   data->speed = ntohs(data->speed);
   data->bearing = ntohs(data->bearing);
   data->sec = ntohs(data->sec);
   data->geoidSep = ntohs(data->geoidSep);
   data->PDOP = ntohs(data->PDOP);
   data->HDOP = ntohs(data->HDOP);
   data->VDOP = ntohs(data->VDOP);
}


int main(int argc, char *argv[]) {
   int fd;
   GPS_DATA g;
   unsigned char initval[] = {0x03,0x37,0xF9,0x80,0x00,0x00,0x00,0x00,0x00,0x00,0x05,0x5A,0x00,0x00,0x00,0x00,0x00,0x50,0x01,0x06,0x00,0x1A,0xB2,0x25,0x00,0x82,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

   unsigned long int val = 54000000;
   int i;

   /*
   fd = GPS_Init();
   if (fd == -1) {
      return 0;
   }
   */

   for (i=0; i<4; i++) {
      printf("0x%02X ", ((unsigned char*)&val)[i]);
   }
   printf("\n");
   printf("GPS_DATA size = %d\n", sizeof(GPS_DATA));
   printf("latitude: %d\n", (int)&(g.latitude) - (int)&g);
   printf("longitude: %d\n", (int)&(g.longitude) - (int)&g);
   printf("altitude: %d\n", (int)&(g.altitude) - (int)&g);
   printf("speed: %d\n", (int)&(g.speed) - (int)&g);
   printf("bearing: %d\n", (int)&(g.bearing) - (int)&g);
   printf("numSV: %d\n", (int)&(g.numSV) - (int)&g);
   printf("year: %d\n", (int)&(g.year) - (int)&g);
   printf("month: %d\n", (int)&(g.day) - (int)&g);
   printf("day: %d\n", (int)&(g.day) - (int)&g);
   printf("hr: %d\n", (int)&(g.hr) - (int)&g);
   printf("min: %d\n", (int)&(g.min) - (int)&g);
   printf("sec: %d\n", (int)&(g.sec) - (int)&g);
   printf("geoidSep: %d\n", (int)&(g.geoidSep) - (int)&g);
   printf("PDOP: %d\n", (int)&(g.PDOP) - (int)&g);
   printf("HDOP: %d\n", (int)&(g.HDOP) - (int)&g);
   printf("VDOP: %d\n", (int)&(g.VDOP) - (int)&g);
   printf("fix: %d\n", (int)&(g.fix) - (int)&g);
   printf("navMode: %d\n", (int)&(g.navMode) - (int)&g);

   /*
   while (1) {
      GPS_Receive(fd);   
      GPS_Translate();
   }
   */
   for (i=0; i<34; i++) {
      ((unsigned char*)&g)[i] = initval[i];
   }
   printf("\nInitially loaded values:\n");
   OutputGPS(&g);

   ConvertGPS(&g);

   printf("\nConverted values:\n");
   OutputGPS(&g);


   close(fd);

   return 0;
}

// outputs all of the stored GPS data values
void OutputGPS(const GPS_DATA *gpsData) {
   // GPRMC
   printf("Time - %02d:%02d:%06.3f\n", gpsData->hr, gpsData->min, ((double)gpsData->sec)/1000.0);
   printf("Latitude  - %10ld min - %13.6f deg\n", gpsData->latitude, ((double)gpsData->latitude)/600000.0);
   printf("Longitude - %10ld min - %13.6f deg\n", gpsData->longitude, ((double)gpsData->longitude)/600000.0);
   printf("Speed - %6d - %8.2f knots\n", gpsData->speed, ((double)gpsData->speed)/100.0);
   printf("Bearing - %6d - %8.2f deg\n", gpsData->bearing, ((double)gpsData->bearing)/100.0);
   printf("Date - %02d/%02d/%02d\n", gpsData->month, gpsData->day, gpsData->year);
   printf("Nav mode - %2d\n", gpsData->navMode);

   // GPGGA
   printf("Num SV - %2d\n", gpsData->numSV);
   printf("Altitude    - %8ld - %8.1f\n", gpsData->altitude, ((double)gpsData->altitude)/10.0);
   printf("Geoidal Sep - %8d - %8.1f\n", gpsData->geoidSep, ((double)gpsData->geoidSep)/10.0);

   // GPGSA
   printf("Fix  - %4dD\n", gpsData->fix);
   printf("PDOP - %8d - %6.2f\n", gpsData->PDOP, ((double)gpsData->PDOP)/100.0);
   printf("HDOP - %8d - %6.2f\n", gpsData->HDOP, ((double)gpsData->HDOP)/100.0);
   printf("VDOP - %8d - %6.2f\n", gpsData->VDOP, ((double)gpsData->VDOP)/100.0);
}
