/****** gpslog.c ******/

/* Outputs a human readable version of a binary data log on a SD card
*/


#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <netinet/in.h>
#include <math.h>


typedef struct {
   long latitude;          // min * 10000, N = +, S = -
   long longitude;         // min * 10000, E = +, W = -
   unsigned long altitude; // m * 10
   unsigned short speed;   // knot * 100
   unsigned short bearing; // deg * 100 (0-360.00)
   unsigned char numSV;    // number of SV's used in solution
   unsigned char year;     // 00-99
   unsigned char month;    // 00-12
   unsigned char day;      // 00-31
   unsigned char hr;       // 00-23
   unsigned char min;      // 00-59
   unsigned short sec;     // s * 1000, 0-59000
   short geoidSep;         // m * 10, geoidal separation
   short PDOP, HDOP, VDOP; // DOP * 10 values
   unsigned char fix;      // 0 = no fix, 2 = 2D, 3 = 3D   
   unsigned char navMode;  // 0 = no fix, 1 = Autonomous mode, 2 = Differential mode, 3 = DR
} GPS_DATA;



// outputs all of the stored GPS data values
void OutputGPS(const GPS_DATA *gpsData);

// converts the endianness of incoming data from the microcontroller
// The micro uses big endian, but linux uses little endian
void ConvertGPS(GPS_DATA *data);


#define BLOCK_SIZE 512

// the size of the struct as it is stored by the micro
#define GPS_SIZE_MICRO 54
// used to be 34

// multiplied by speed in knots to get degrees of lat
// such that 1ft vector == 5mph
const double vecScale = 6.301795e-7; 

#define DEG2RAD 0.0174532925



int main(int argc, char* argv[])
{  
   int blocknum, address;
   char *filefrom, *fileto;
   FILE *fpin, *fpout;
   unsigned char buf[BLOCK_SIZE];
   GPS_DATA gps;
   int i, j, k, blockCount, totalcount=0, storedcount=0;
   int SV[30];
   
   double dlat, dlong, delev;
   
   unsigned short radio[6];
   unsigned short therm[4];


   if (argc != 4) {
      printf("Usage: %s <input> blocknum <output>\n", argv[0]);
      return 0;
   }

   filefrom = argv[1];
   blocknum = atoi(argv[2]);      
   fileto = argv[3]; 
   
   address = blocknum * BLOCK_SIZE;
   for (i=0; i<30; i++) SV[i] = 0;


   if ((fpin = fopen(filefrom, "rb"))) {
      if ((fpout = fopen(fileto, "w"))) {
         // move to proper offset
         fseek(fpin, address, SEEK_SET);

         // the first 4 bytes contains the number of blocks stored
         fread(buf, BLOCK_SIZE, 1, fpin);
         blockCount = ntohl(*(unsigned long int *)buf);
         printf("Block count: %d\n", blockCount);

         //fprintf(fpout, "lat,long,elev,time,SV,fix,speed_mph,bearing,radio0,radio1,radio2,radio3,radio4,radio5,therm0,therm1,therm2,therm3\n");
			
         // read in all of the blocks
         for (i=0; i<blockCount; i++) {            
            fread(buf, BLOCK_SIZE, 1, fpin);

            // read all of the entries in each block
            for (j=0; j+GPS_SIZE_MICRO<=BLOCK_SIZE; j+=GPS_SIZE_MICRO) {
                              
               gps = *(GPS_DATA *)(&buf[j+20]);
               ConvertGPS(&gps);
               totalcount++;
               
               dlat = ((double)gps.latitude)/600000.0;
               dlong = ((double)gps.longitude)/600000.0; 
               delev = ((double)gps.altitude)/10.0; 
               
               if (gps.fix > 2 && dlat > 35. && dlat < 45. && dlong > -89. && dlong < -85 && delev>40. && delev<500) {	  
                  storedcount++;           
                                 
                  for (k=0; k<6; k++){
                     radio[k] = ntohs(*(unsigned short *)(&buf[j + 2*k]));
                  }
                  
                  for (k=0; k<4; k++){
                     therm[k] = ntohs(*(unsigned short *)(&buf[j + 2*k + 12]));
                  }
                  
                  
                  // output waypoint
                  fprintf(fpout, "%.7f,%.7f,", ((double)gps.latitude)/600000.0, ((double)gps.longitude)/600000.0);
                  fprintf(fpout, "%.1f,", ((double)gps.altitude)/10.0);
                  fprintf(fpout, "%04d,%02d,%02d,%02d,%02d,%06.3f,", gps.year+2000, gps.month, gps.day, gps.hr, gps.min, ((double)gps.sec)/1000.0);                  
                  fprintf(fpout, "%d,", gps.numSV);
                  fprintf(fpout, "%d,", gps.fix);
                  fprintf(fpout, "%.2f,", ((double)gps.speed)/100.0 * 1.15078); //mph
                  fprintf(fpout, "%.2f,", ((double)gps.bearing)/100.0);
      				
      				for (k=0; k<6; k++) {
      				   fprintf(fpout, "%d,", radio[k]);   
      				}
      				
      				for (k=0; k<4; k++) {
      				   fprintf(fpout, "%d", therm[k]);
      				   if (k<3) fprintf(fpout, ",");  
      				}
                  
                  fprintf(fpout, "\n");
               }
               
               SV[gps.numSV]++;
				                 
            }
                                    
         }
                  
         fclose(fpout);
      }
      else {
         printf("Could not open: %s, for writing\n", fileto);
      }
      fclose(fpin);
   }
   else {
      printf("Could not open: %s, for reading\n", filefrom);
   }
   
   printf("\nSV Counts:\n");
   for (i=0; i<13; i++) {
      printf(" %2d - %5d\n", i, SV[i]);
   }

   printf("\nTotal entries: %d\nEntries Stored: %d\n", totalcount, storedcount);
	return 0;
}



// outputs all of the stored GPS data values
void OutputGPS(const GPS_DATA *gpsData) {
   // GPRMC
   printf("Time - %02d:%02d:%06.3f\n", gpsData->hr, gpsData->min, ((double)gpsData->sec)/1000.0);
   printf("Latitude  - %10ld min - %13.6f deg\n", gpsData->latitude, ((double)gpsData->latitude)/600000.0);
   printf("Longitude - %10ld min - %13.6f deg\n", gpsData->longitude, ((double)gpsData->longitude)/600000.0);
   printf("Speed - %6d - %8.2f knots\n", gpsData->speed, ((double)gpsData->speed)/100.0);
   printf("Bearing - %6d - %8.2f deg\n", gpsData->bearing, ((double)gpsData->bearing)/100.0);
   printf("Date - %02d/%02d/%02d\n", gpsData->month, gpsData->day, gpsData->year);
   printf("Nav mode - %2d\n", gpsData->navMode);

   // GPGGA
   printf("Num SV - %2d\n", gpsData->numSV);
   printf("Altitude    - %8ld - %8.1f\n", gpsData->altitude, ((double)gpsData->altitude)/10.0);
   printf("Geoidal Sep - %8d - %8.1f\n", gpsData->geoidSep, ((double)gpsData->geoidSep)/10.0);

   // GPGSA
   printf("Fix  - %4dD\n", gpsData->fix);
   printf("PDOP - %8d - %6.2f\n", gpsData->PDOP, ((double)gpsData->PDOP)/100.0);
   printf("HDOP - %8d - %6.2f\n", gpsData->HDOP, ((double)gpsData->HDOP)/100.0);
   printf("VDOP - %8d - %6.2f\n", gpsData->VDOP, ((double)gpsData->VDOP)/100.0);
}


// converts the endianness of incoming data from the microcontroller
// The micro uses big endian, but linux uses little endian
void ConvertGPS(GPS_DATA *data) {
   data->latitude = ntohl(data->latitude);
   data->longitude = ntohl(data->longitude);
   data->altitude = ntohl(data->altitude);
   data->speed = ntohs(data->speed);
   data->bearing = ntohs(data->bearing);
   data->sec = ntohs(data->sec);
   data->geoidSep = ntohs(data->geoidSep);
   data->PDOP = ntohs(data->PDOP);
   data->HDOP = ntohs(data->HDOP);
   data->VDOP = ntohs(data->VDOP);
}
