/******************************* nav.c *******************************/


// Calculates the distance and bearing between two GPS waypoints
// using only integer math


#include <stdio.h> 
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <netinet/in.h>

#include "nav.h"


// cos(ang) * 256 with ang in degree
// used for scaling down longitude values based on latitude
// cosAdjust[0] corresponds to 4 deg as 0-3deg == 256
const unsigned char cosAdjust[87] = {
   255, 255, 255, 254, 254, 253, 252, 251, 250, 249, 248, 247, 246, 245, 243,
   242, 241, 239, 237, 236, 234, 232, 230, 228, 226, 224, 222, 219, 217, 215,
   212, 210, 207, 204, 202, 199, 196, 193, 190, 187, 184, 181, 178, 175, 171,
   168, 165, 161, 158, 154, 150, 147, 143, 139, 136, 132, 128, 124, 120, 116,
   112, 108, 104, 100,  96,  92,  88,  83,  79,  75,  71,  66,  62,  58,  53,
    49,  44,  40,  36,  31,  27,  22,  18,  13,   9,   4,   0};


// calculate the square root of a long input value
// source: http://www.codecodex.com/wiki/index.php?title=Calculate_an_integer_square_root
unsigned long lsqrt(unsigned long y);


// calculates the distance (ft) and bearing (deg) between two waypoints
// bearing vector goes from p1 to p2

void Nav_CalcDifference(unsigned long *pDist, unsigned long *pBearing, const GPS_DATA *p1, const GPS_DATA *p2) {
   long dlat, dlong, temp;
   int downshift=0;
   unsigned char target, type, low=0, mid=43, high=86;
   unsigned long udlat, udlong, utemp, utemp2, uval;

   //printf("p1: (%ld, %ld), p2: (%ld, %ld)\n", p1->latitude, p1->longitude, p2->latitude, p2->longitude);
   // calculate the differences between the waypoints
   // differences currently in units of minutes of angle
   dlat = p2->latitude - p1->latitude;
   dlong = p2->longitude - p1->longitude;

   // convert from minute to degree, rounding to the nearest degree
   if (p1->latitude > 0) temp = p1->latitude / 600000L;
   else temp = p1->latitude / -600000L;

   // scale dlong as longitude degrees get closer together at higher latitudes
   // no change for latitudes <4deg
   if (temp > 90) {
      dlong = 0;      
   }
   else if (temp > 3) {
      dlong *= cosAdjust[temp - 4];
      dlong >>= 8;
   }

   // absolute values
   udlat = (dlat > 0 ? dlat : -dlat);
   udlong = (dlong > 0 ? dlong : -dlong);

   //printf("dlat:  %ld\ndlong: %ld\n", dlat, dlong);

   // downshift the inputs so that the square root doesn't fail
   while (udlat > 0x8000 || udlong > 0x8000) {
      udlat >>= 1;
      udlong >>= 1;
      downshift++;
   }

   //printf("Downshift=%d, udlat=%ld, udlong=%ld\n", downshift, udlat, udlong);

   // calculate distance
   *pDist = udlat * udlat + udlong * udlong;
   *pDist = lsqrt(*pDist);

   // scale back up
   // pDist is units of minutes * 10000
   *pDist <<= downshift;

   //printf("Distance: %ld (min)\n", *pDist);
   


   // absolute values
   udlat = (dlat > 0 ? dlat : -dlat);
   udlong = (dlong > 0 ? dlong : -dlong);

   utemp = *pDist;
   downshift = 0;

   // pick the triangle leg such that sine or cosine is small
   if (udlat < udlong) { 
      uval = udlat;
      type = 0;
   }
   else {
      uval = udlong;
      type = 1;
   }

   //printf("uval: %ld\n", uval);;

   // ensure it doesn't go out of range
   while (uval > 0x00FFFFFFuL) {
      uval >>= 1;
      utemp >>= 1;
      downshift++;
   }

   // calulate the target sine or cosine
   utemp2 = uval << 8; 
   //printf("uval: %ld  utemp2: %ld\n", uval, utemp2);
   utemp2 /= utemp;
   //printf("utemp: %ld  utemp2: %ld\n", utemp, utemp2);
   target = (unsigned char)utemp2;
   //printf("Target: %d\n", target);

   // convert from minute * 10000 to feet
   *pDist = (utemp * 155) >> (8 - downshift);
   //printf("Distance: %ld (ft)\n", *pDist);


   //printf("Type: %d\n", type);
   if (uval == 0) {
      // corner case of vertical or horizontal travel
      mid = 90;
   }
   else {

      // find the angle using bisection   
      while (low < mid && mid < high) {
         //printf("L: %2d, M: %2d, H: %2d\n", low, mid, high);
         if (target == cosAdjust[mid]) {
            break;
         }
         else if (target < cosAdjust[mid]) {
            low = mid;
         }
         else {
            high = mid;
         }
         mid = (high + low) >> 1;      
      }

      // make mid an angle away from north
      mid += 4;
   }

   if (type) mid = 90 - mid;

   //printf("Angle: %d\n", mid);

   if (dlat >= 0 && dlong >= 0) {
      // quandrant I
      *pBearing = mid;
   }
   else if (dlat >= 0 && dlong <= 0) {
      // quandrant II
      *pBearing = 360 - mid;
   }
   else if (dlat <= 0 && dlong <= 0) {
      // quandrant III
      *pBearing = mid + 180;
   }
   else {
      // quandrant IV
      *pBearing = 180 - mid;
   }

}



int main(int argc, char *argv[]) {
   unsigned long dist, bearing, truedist, truebearing;
   GPS_DATA p1, p2;

   int i, val;
   double dlat, dlong, ddist;


   //printf("Test cases:\n");


   p1.latitude = 24279031L;
   p1.longitude = -52154981L;
   p2.latitude = 24280528L;
   p2.longitude = -52155932L;
   truedist = 1010;
   truebearing = 334;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   //printf("(%f, %f) Real: %ld(ft), %ld(deg), Calc: %ld(ft),  %ld(deg)\n",
   //   p1.latitude / 600000., p1.longitude / 600000., truedist, truebearing, dist, bearing);
   
   truebearing = 154;
   
   Nav_CalcDifference(&dist, &bearing, &p2, &p1);

   //printf("(%f, %f) Real: %ld(ft), %ld(deg), Calc: %ld(ft),  %ld(deg)\n",
   //   p2.latitude / 600000., p2.longitude / 600000., truedist, truebearing, dist, bearing);


   p1.latitude = 24280008L;
   p1.longitude = -52155748L;
   p2.latitude = 24280029L;
   p2.longitude = -52155247L;
   truedist = 233;
   truebearing = 86;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   //printf("(%f, %f) Real: %ld(ft), %ld(deg), Calc: %ld(ft),  %ld(deg)\n",
   //   p1.latitude / 600000., p1.longitude / 600000., truedist, truebearing, dist, bearing);

   truebearing = 266;
   
   Nav_CalcDifference(&dist, &bearing, &p2, &p1);

   //printf("(%f, %f) Real: %ld(ft), %ld(deg), Calc: %ld(ft),  %ld(deg)\n",
   //   p2.latitude / 600000., p2.longitude / 600000., truedist, truebearing, dist, bearing);


   p1.latitude = 24256028L;
   p1.longitude = -52141495L;
   p2.latitude = 24255385L;
   p2.longitude = -52154478L;
   truedist = 6036;
   truebearing = 267;
   
   Nav_CalcDifference(&dist, &bearing, &p1, &p2);

   //printf("(%f, %f) Real: %ld(ft), %ld(deg), Calc: %ld(ft),  %ld(deg)\n",
   //   p1.latitude / 600000., p1.longitude / 600000., truedist, truebearing, dist, bearing);



   // circle test case
   ddist = 100.;
   //ddist = 1056000.;

   p1.latitude = 24256028L;
   p1.longitude = -52141495L;

   for (i=0; i<360; i++) {
      dlat = ddist * sin(i * 3.1415926535 / 180.0);
      dlong = ddist * cos(i * 3.1415926535 / 180.0);

      dlong /= cos(p1.latitude / 600000.0 * 3.1415926535 / 180.0);

      ////printf("%3d %f %f\n", i, dlat, dlong);

      p2.latitude = p1.latitude + dlat/0.60870455;
      p2.longitude = p1.longitude + dlong/0.60870455;

      Nav_CalcDifference(&dist, &bearing, &p1, &p2);

      // convert to bearing
      val = 90 - i;
      if (val < 0) val += 360;

      printf("%3d\t%ld\t%ld\n", val, dist, bearing);

   }

   return 0;
}





// calculate the square root of a long input value
// source: http://www.codecodex.com/wiki/index.php?title=Calculate_an_integer_square_root
unsigned long lsqrt(unsigned long y)
{
   unsigned long x_old, x_new;
   unsigned long testy;
   int nbits;
   int i;

   if (y == 0) {
      return 0L;
   }

   // select a good starting value using binary logarithms: 
   nbits = (sizeof(y) * 8);
   for (i = 4, testy = 16L;; i += 2, testy <<= 2L) {
      if (i >= nbits || y <= testy) {
         x_old = (1L << (i >> 1));
         // x_old = closest power of 2 to sqrt(testy) 
         break;
      }
   }
   
   // x_old >= sqrt(y)
   // use the Babylonian method to arrive at the integer square root:
   // i should never be very big, but set a cap just in case

   for (i=0;i<50;i++) { 
      x_new = (y / x_old + x_old) >> 1;
      if (x_old <= x_new)
            break;
      x_old = x_new;
   }   
   
   // make sure that the answer is right: 
   /*
   if (x_old * x_old > y || (x_old + 1) * (x_old + 1) <= y) {
      //printf("Error in lsqrt(), bad output\n");
      return 0L;
   }
   */
   
   return x_old;
}
